local SAFE_MODE = true

local lfg

do
    local _G = getfenv(0)
    
    lfg = _G["Grouper-1.0"] or error("Grouper-1.0 core is missing.")
    
    -- inject Grouper Core into AceAddon
    local lfg2 = AceLibrary and AceLibrary("AceAddon-2.0"):new("AceConsole-2.0","AceEvent-2.0",
        "AceDebug-2.0","AceDB-2.0","FuBarPlugin-2.0","AceHook-2.1") or error("AceLibrary is required for GrouperLFG Ace2 bindings")
    
    for k,v in pairs(lfg) do
        lfg2[k] = v
    end
    
    _G["GrouperLFG-1.0"] = lfg2
    lfg = lfg2
end

local tourist = AceLibrary and AceLibrary("Tourist-2.0") or error("Library Tourist-2.0 is required")
local quixote = AceLibrary and AceLibrary("Quixote-1.0") or error("Library Quixote-1.0 is required")
local tablet = AceLibrary and AceLibrary("Tablet-2.0") or error("Library Tablet-2.0 is required") -- when I get around to tooltips
local Bits = AceLibrary and AceLibrary("Bits-1.0") or Bits or error("Library Bits-1.0 is required")
local dewdrop = AceLibrary and AceLibrary("Dewdrop-2.0")

local mathmin, mathmax = math.min, math.max

-- =============================================
--                  CONSTANTS
-- =============================================

table.insert(lfg.PRIORITY["Current Zone's Quests"],"Quixote_Update")
table.insert(lfg.PRIORITY["Adjacent Zones' Quests"],"Quixote_Update")
table.insert(lfg.PRIORITY["All Quests"],"Quixote_Update")
table.insert(lfg.PRIORITY["Select Quest..."],"Quixote_Update")

local MENU_PRI_LOOKING = "Looking for a group in this category."
local MENU_PRI_NOT_LOOKING = "Not looking for a group in this category."
local MENU_PRI_NONE = "No objectives available in this category."
local MENU_PRI_FILTERED = "No unfiltered objectives available in this category."

-- TODO REMOVE THESE LOCAL FUNCS
local colorBits = Bits:new(lfg:GetColors())
local sizeBits = Bits:new(lfg:GetInstanceSizes())
local continentBits = Bits:new(lfg:GetContinents())
local temp

local cGREY = "|cFF808080"
local cRED = "|cFFFF0000"
local cORANGE = "|cFFFFCC00"
local cTRANS = "|c00000000"
local cRESET = "|r"

local stateBits = Bits:new(
    -- primary state
    "DND",          -- player selected DND
    "AVAILABLE",    -- player selected available
    "JOIN",         -- player selected join
    "FORM",         -- player selected form
    
    "AUTO",         -- auto upgrade/downgrade of state
    "TAXI",         -- player on a taxi, will transition to primary state
    "GROUPED",      -- player is grouped, will transition to available or dnd
    "GROUPEDOK",    -- modifies grouped to permit available
    "LEADER",       -- player is group leader
    "AFK"           -- player is <AFK> or <DND> and preferences recognize these
)

-- player logged in, will transition to available or dnd per preferences
stateBits:Alias("LOGIN","AUTO","DND")
-- join didn't find a group, attempting to form, (if joining a group transitions to available grouped)
stateBits:Alias("FORM AUTO","AUTO","FORM")
stateBits:Alias("PRIMARY","DND","AVAILABLE","JOIN","FORM")

local STATE_VALUE = stateBits:ReverseMap()

local RGBtoHSB
do
    local HRED = 2 * math.pi
    local BLUE = 4 * math.pi / 3
    local GREEN = 2 * math.pi / 3
    local ARC_60 = math.pi / 3
    -- Algorithm from "Computer Graphics -- Principles and Practice,"
    -- Foley et al, 1996, pp. 592-593
    function RGBtoHSB(r,g,b)
        local brightness = mathmax(r,g,b)
        local delta = brightness - mathmin(r,g,b)
        local saturation = brightness == 0 and 0 or (delta / brightness)
        
        if saturation == 0 then
            return 0, saturation, brightness
        else
            if r == brightness then
                if b > g then
                    return (HRED + (ARC_60 * g - ARC_60 * b) / delta), saturation, brightness
                else
                    return ((ARC_60 * g - ARC_60 * b) / delta), saturation, brightness
                end
            elseif g == brightness then
                return (GREEN + (ARC_60 * b - ARC_60 * r) / delta), saturation, brightness
            else
                return (BLUE + (ARC_60 * r - ARC_60 * g) / delta), saturation, brightness
            end
        end
    end
end

local GetColorWord
do
    local RED = 0
    local ORANGE = RGBtoHSB(1,0.5,0)
    local YELLOW = RGBtoHSB(0.5,1,0)
    local _,GREY = RGBtoHSB(0.50,0.50,0.50)

    function GetColorWord(r,g,b)
        local h,s,b = RGBtoHSB(r,g,b)
        if s < 1 then
            return b==1 and "WHITE" or "GREY"
        elseif h > YELLOW then
            return "GREEN"
        elseif h > ORANGE then
            return "YELLOW"
        elseif h > RED then
            return "ORANGE"
        else
            return "RED"
        end
    end
end

-- =============================================
--             DATABASE & ACCESSORS
-- =============================================

local NAME = "Grouper"

local bitsCount = colorBits:Size() + sizeBits:Size() + continentBits:Size()

lfg:RegisterDB(NAME.."DB",NAME.."DBPC")

lfg:RegisterDefaults('char', {
    filter = { -- filter is per-char, it depends on strength of your character
        zoneColor = colorBits:GetHash("YELLOW","GREEN"),
        instanceColor = colorBits:GetHash("ORANGE","YELLOW","GREEN"),
        questColor = colorBits:GetHash("RED","ORANGE","YELLOW","GREEN"),
        instanceSize = sizeBits:GetHash(5,10),
        continent = continentBits:GetHash(lfg:GetContinents()),
    },
    -- priority is per-char, it depends on your character's goals
    priority = {unpack((select(3,lfg:GetLFGProfile("Default"))))},
    note = select(4,lfg:GetLFGProfile("Default")),
    selections = {},
    -- double-check serialversion, users love to play with saved variables
    serialversioncheck = bitsCount,
})

-- TODO if any priorities absent, add them to self.db.char.priority at the end

lfg:RegisterDefaults('profile', {
    option = { -- option is profile, it depends on behavior
        state = "AVAILABLE",    -- default state to start in DND|AVAILABLE
        loginPause = 2,       -- seconds to pause on login 0+ : 0 --> go to avail right away
        ignoreAFK = false,      -- do not use afk restrictions
        ignoreDND = false,      -- do not use dnd restrictions
        ignoreTaxi = false,     -- do not use taxi restrictions (ADVANCED)
        availableInGroup = false,    -- do not use group restrictions (ADVANCED)
        autoForm = true,        -- do / don't auto-form group after looking fails
        autoFormPause = 5,    -- seconds to pause to upgrade join to form 60+
    },
    serialversion = { -- will be used to upgrade the data values if they change
        check = bitsCount,
        colorSet = {lfg:GetColors()},
        sizeSet = {lfg:GetInstanceSizes()},
        continentSet = {lfg:GetContinents()},
    },
})

function lfg:GetObjectiveSelection(priority,objective)
    if not self.db.char.selections[priority] then
        self:Debug("created selection list",priority)
        self.db.char.selections[priority] = {}
    end
    
    self:Debug("GetObjectiveSelection",priority,objective,"==",
        self.db.char.selections[priority][objective])
    return self.db.char.selections[priority][objective]
end

function lfg:ToggleObjectiveSelection(priority,objective,value)
    self:Debug("ToggleObjectiveSelection",priority,objective,"=",value)
    if not self.db.char.selections[priority] then
        self:Debug("created selection list",priority)
        self.db.char.selections[priority] = {}
    end
    
    self.db.char.selections[priority][objective] = value
    self:UpdateAutoLFG()
end

function lfg:ListObjectiveSelection(priority)
    if not self.db.char.selections[priority] then
        return
    end
    
    return unpack(self.db.char.selections[priority])
end

function lfg:GetDB()
    return self.db
end

function lfg:GetZoneColorFilter()
    return self.db.char.filter.zoneColor
end

function lfg:SetZoneColorFilter(value)
    self.db.char.filter.zoneColor = value
    self:UpdateFilters()
end

function lfg:GetInstanceColorFilter()
    return self.db.char.filter.instanceColor
end

function lfg:SetInstanceColorFilter(value)
    self.db.char.filter.instanceColor = value
    self:UpdateFilters()
end

function lfg:GetQuestColorFilter()
    return self.db.char.filter.questColor
end

function lfg:SetQuestColorFilter(value)
    self.db.char.filter.questColor = value
    self:UpdateFilters()
end

function lfg:GetInstanceSizeFilter()
    return self.db.char.filter.instanceSize
end

function lfg:SetInstanceSizeFilter(value)
    self.db.char.filter.instanceSize = value
    self:UpdateFilters()
end

function lfg:GetContinentFilter()
    return self.db.char.filter.continent
end

function lfg:SetContinentFilter(value)
    self.db.char.filter.continent = value
    self:UpdateFilters()
end

function lfg:IsIgnoreAFK()
    return self.db.profile.option.ignoreAFK
end

function lfg:ToggleIgnoreAFK()
    self.db.profile.option.ignoreAFK = not self.db.profile.option.ignoreAFK
end

function lfg:IsIgnoreDND()
    return self.db.profile.option.ignoreDND
end

function lfg:ToggleIgnoreDND()
    self.db.profile.option.ignoreDND = not self.db.profile.option.ignoreDND
end

function lfg:IsIgnoreTaxi()
    return self.db.profile.option.ignoreTaxi
end

function lfg:ToggleIgnoreTaxi()
    self.db.profile.option.ignoreTaxi = not self.db.profile.option.ignoreTaxi
end

function lfg:IsAvailableInGroup()
    return self.db.profile.option.availableInGroup
end

function lfg:ToggleAvailableInGroup()
    self.db.profile.option.availableInGroup = not self.db.profile.option.availableInGroup
end

function lfg:GetLoginPause()
    return self.db.profile.option.loginPause
end

function lfg:SetLoginPause(value)
    self.db.profile.option.loginPause = value
end

function lfg:IsAutoForm()
    return self.db.profile.option.autoForm
end

function lfg:ToggleAutoForm()
    self.db.profile.option.autoForm = not self.db.profile.option.autoForm
end

function lfg:GetAutoFormPause()
    return self.db.profile.option.autoFormPause
end

function lfg:SetAutoFormPause(value)
    self.db.profile.option.autoFormPause = value
end

function lfg:GetInitialState()
    return self.db.profile.option.state
end

function lfg:SetInitialState(state)
    self.db.profile.option.state = state
end

function lfg:GetPriorities()
    return self.db.char.priority
end

function lfg:LoadLFGProfile(name)
    local i, key, order, note = self:GetLFGProfile(name)
    local pri = {}
    for i,v in ipairs(order) do
        table.insert(pri,v)
    end
    self.db.char.priority = pri
    self.db.char.note = note
    
    self:UpdateAllPriorityOrder()
end
-- =============================================
--             FUBAR FUNCTIONS
-- =============================================

local TOOLTIP_HINT = "Click to process LFGQuery (if flashing)\n"..
    "Shift-click to toggle LFG\nCtrl-click to form a group\n\nKey:LFG:State "..
    "( #in queue / #potential / #total LFG )"
    
function lfg:OnTooltipUpdate()
    tablet:SetHint(TOOLTIP_HINT)
end

local OnLFGUpdate, populateBitsSubmenu, Bar_OnCategoryCountUpdate, Bar_OnObjectiveCountUpdate,
    Bar_OnCategoryUpdate, Bar_ExecuteLFG, Bar_OnLFGUpdate, Bar_OnObjectivesUpdate

lfg.hasIcon = "Icon\\blue check.tga"

local function separator(order,name)
    return {
        type = 'header',
        name = name or cTRANS..".",
        order = order,
    }
end

function newGetBits(bits,get,handler)
    if type(get) == "text" then
        get = handler[get]
    end
    
    local function getFunc(key) return bits:HashContains(get(handler),key) end
    
    return getFunc
end

function newSetBits(bits,get,set,handler)
    if type(get) == "text" then
        get = handler[get]
    end
    
    if type(set) == "text" then
        set = handler[set]
    end
    
    local function setFunc(key,value)
        local bitSetFunction = value and bits["AddToHash"] or bits["RemoveFromHash"]
        return set(handler,bitSetFunction(bits,get(handler),key))
    end
    
    return setFunc
end

function populateBitsSubmenu(bitsSubmenu,bits,get,set,handler,...)
    bitsSubmenu.type = 'group'
    bitsSubmenu.pass = true
    bitsSubmenu.args = {}
    local getbit = newGetBits(bits,get,handler)
    bitsSubmenu.get = getbit
    assert(getbit, "get bits submenu is nil")
    local setbit = newSetBits(bits,get,set,handler)
    bitsSubmenu.set = setbit
    assert(setbit, "set bits submenu is nil")
    
    for i=1,select('#',...) do
        local v = select(i,...) -- local v must be here for proper get/set closure
--[[        local p,val = pcall(getbit,v)
        assert(p,"get bits function failed: "..tostring(val))
        assert(val ~= nil,"get bits function retunred nil")
        p,val = pcall(setbit,v,val)
        assert(p,"set bits function failed: "..tostring(val))]]
        
        bitsSubmenu.args[v] = {
            type = 'toggle',
            name = tostring(v),
            desc = "Toggle "..v,
            order = i,
        }
    end
    
    return bitsSubmenu
end

function Bar_OnCategoryCountUpdate(self,catNum)
    if not catNum then return end
    local catText = self:GetCategoryText(catNum)
    local text = cORANGE.."("..self:GetCategoryStatus(catNum)..")  "..cRESET..catText
    self.browse.args[catText].name = text
    self:Debug(catText,"name set",text)
end

function Bar_OnObjectiveCountUpdate(self,categoryText,query)
    if not query then return end
    local text = cORANGE.."("..self:GetObjectiveStatus(query)..")  "..cRESET..self:GetObjectiveText(query)
    self.browse.args[categoryText].args[query].name = text
    self:Debug(categoryText,query,"name set",text)
end

function Bar_OnCategoryUpdate(self,newcategories)
    self:Debug("Bar_OnCategoryUpdate")
    
    local cats = self.browse.args
    
    for i,v in ipairs(newcategories) do
        local category = cats[v]
        if not category then
            self:Debug("add bar category ",v)
            category = {
                type = 'group',
                desc = "Browse LFG for a "..v,
                args = {},
            }
            cats[v] = category
        end
        
        category.order = i
        Bar_OnCategoryCountUpdate(self,i)
    end
end

local busyMenu = {
    type = 'header',
    name = "Updating...",
    order = 1,
}

function Bar_ExecuteLFG(self,query)
    self:Debug("UPDATING query",self:GetQueryText(query))
    
--    local now = time()
--    local cache = self.cache[query]
    
    self:ProcessQuery(query)
    -- TODO make category args "processing"
end

function Bar_OnLFGUpdate(self,categoryText,query)
    self:Debug("Bar_OnLFGUpdate")
    
    local cache = self.cache[query]
    if not cache then
        self:Debug("cache empty",query)
        return
    end
    
    local objective = self.browse.args[categoryText].args[query]
    if objective then
        local args = objective.args
        
        if not args.space then
            args.space = {
                type = 'header',
                order = 2,
            }
            args[1] = {}
        end
        
        local i,argn,cachen = 1,#objective.args,#cache-1
        
--        objective.name = cORANGE.."(".. cachen ..")  "..cRESET..self:GetObjectiveText(query)
        
        while i <= argn or i <= cachen do
            local item = args[i]
            local name = cache[i+1] -- needed for closure
            if not item then
                item = {
                    type = 'header',
                    order = i+2,
                    hidden = true,
                }
                args[i] = item
            end
            
            if name then
                local time, level, zone, class, criteria1,
                    criteria2, criteria3, comment, numParty, isLFM = self:GetWhoData(name)
                
                item.type = 'execute'
                -- TODO get better who data formatting
                if not time then
                    item.name = name
                    item.desc = "no who data"
                else
                    item.name = string.format("%s %s %s%s%s",name,level,class,
                        isLFM == "true" and " LFM" or "",
                        tonumber(numParty) > 0 and " ("..(numParty+1).."/"..
                            self:GetGroupSize(query,categoryText,nil)..")" or ""
                        )
                    item.desc = string.format("%s\nLooking For Group:\n%s%s%s%s%s%s%s",zone,comment,
                        string.len(comment) > 0 and "\n" or "",criteria1,
                        string.len(criteria1) > 0 and "\n" or "",criteria2,
                        string.len(criteria2) > 0 and "\n" or "",criteria3,
                        string.len(criteria3) > 0 and "\n" or "")
                end
                item.hidden = false
                item.func = function() ChatFrame_SendTell(name) end
            else
                if i == 1 then
                    item.type = 'header'
                    item.name = "None"
                    item.desc = nil
                    item.hidden = false
                    item.func = nil
                else
                    item.hidden = true
                    item.func = nil
                end
            end
            
            i = i + 1
        end
    end
end

function Bar_OnObjectivesUpdate(self)
    self:Debug("Bar_OnObjectivesUpdate")
    
    local objective, catText
    local cats = self.browse.args
    
    for i,v in ipairs(self.categories) do
        cats[v].args = {}
    end
    
    local i = 1
    for query,text in pairs(self.objectives) do
        catText = self:GetQueryText(query)
        self:Debug("add bar objective",query,catText,text)
        objective = cats[catText].args[query]
        if not objective then
            local queryCl = query
            objective = {
                type = 'group',
                args = {
                    update = {
                        type = 'execute',
                        name = "Update",
                        desc = "Query server for players LFG "..catText..":"..text,
                        order = 1,
                        func = function() Bar_ExecuteLFG(self,queryCl) end,
                        -- TODO disable based on query time
                    },
--[[                    space = {
                        type = 'header',
                        order = 2,
                    },
                    [1] = {
                        type = 'header',
                        name = "None",
                        order = 3,
                    },]]
                },
            }
            cats[catText].args[query] = objective
        end
        
        objective.name = cORANGE.."(?)  "..cRESET..text
        objective.desc = "Players LFG for "..catText..":"..text
        objective.order,i = i,i+1
        
        Bar_OnLFGUpdate(self,catText,query)
        Bar_OnObjectiveCountUpdate(self,catText,query)
    end
    
    for i,v in ipairs(self.categories) do
        local show = false
        for _,_ in pairs(cats[v].args) do
            show = true
            break
        end
        if show then
            self:Debug("category shown:",v)
            cats[v].hidden = false
        else
            self:Debug("category hidden:",v)
            cats[v].hidden = true
        end
        Bar_OnCategoryCountUpdate(self,i)
    end
end

function lfg:ShowDebugMenus()
    return not self:IsDebugging()
end

local prioritySubmenuNone = {
    type = 'header',
    name = cGREY.."None",
    order = 7,
}

function lfg:UpdatePriorityCount(element,key,value,numInQueue,
    numVisible,numZonesIgnored,numZonesAvailable,numZonesFiltered,ignored)
    
    if not element then
        element = self.priorities[key]
    end
    
    if not element then return end
    
    local header = cORANGE.."("..numInQueue.."/"..numVisible..")  "
    
    self:Debug("UpdatePriorityCount",element,key,
        value,numInQueue,numVisible,numZonesIgnored,
        numZonesAvailable,numZonesFiltered,ignored)
    
    assert(numZonesAvailable)
    assert(numZonesIgnored)
    
    element.activeCount = numZonesAvailable
    element.passiveCount = numZonesIgnored
    element.filteredCount = numZonesFiltered
    
--    self:PrintLiteral(element)
    
    if numZonesAvailable > 0 then
        element.name = header..cRESET..key
        if ignored then
            element.desc = MENU_PRI_NOT_LOOKING
        else
            element.desc = MENU_PRI_LOOKING
        end
        element.args.none = nil
    else
        if numZonesFiltered > 0 then
            element.desc = MENU_PRI_FILTERED
        else
            element.desc = MENU_PRI_NONE
        end
        element.name = header..key
        if numZonesFiltered < 1 then
            if numZonesIgnored < 1 or numZonesIgnored > 25 then
                element.args.none = prioritySubmenuNone
            else
                element.args.none = nil
            end
        end
    end
    
    self:UpdateAutoLFG()
end

function lfg:UpdateAllPriorityOrder()
    local args = self.priorities
    if not args then return end
    
    local k = self.PRIORITY_IGNORE_HEADER
    local ignoreNum = self:GetPriorityNumber(k)
    self:UpdatePriorityOrder(args[k],k,nil,ignoreNum,ignoreNum)
    
    for k,v in pairs(self.PRIORITY) do
        local num = self:GetPriorityNumber(k)
        self:UpdatePriorityOrder(args[k],k,nil,num,ignoreNum)
    end
end

function lfg:UpdatePriorityOrder(element,key,value,num,ignoreNum)
    if not element then
        element = self.priorities and self.priorities[key]
    end
    
    if not element then return end
    
    if not ignoreNum then
        ignoreNum = self:GetPriorityNumber(self.PRIORITY_IGNORE_HEADER)
    end
    
    if not num then
        num = self:GetPriorityNumber(key)
    end
    
    self:Debug("UpdatePriorityOrder",element,key,value,num,ignoreNum)
    
--    self:PrintLiteral(element)
    
    if element.activeCount and element.activeCount > 0 then
        if ignored then
            element.desc = MENU_PRI_NOT_LOOKING
        else
            element.desc = MENU_PRI_LOOKING
        end
    else
        if element.filteredCount and element.filteredCount > 0 then
            element.desc = MENU_PRI_FILTERED
        else
            element.desc = MENU_PRI_NONE
        end
    end
    
    element.order = num
    
    self:UpdateAutoLFG()
end

function lfg:UpdatePriority(element,key)
    if not element then
        element = self.priorities and self.priorities[key]
    end

    if not element then return end
    
    self:Debug("UpdatePriority",key)
    
    element.updateFunction(self)
end

-- =============================================
--           AUTO-LFG PROFILE SYSTEM
-- =============================================

function lfg:ConstructProfileMenu(profileMenu,order)
    local loadProfileArgs = {}
    for i, key in self:IterateLFGProfiles() do
        loadProfileArgs[key] = {
            type = 'execute',
            name = key,
            desc = "Load "..key.." profile. Changes to your current profile will be lost.",
        }
    end
    
    profileMenu.args.load.args = loadProfileArgs
    profileMenu.args.priorities.args = {}
    
    local moveUp = {
        type = 'execute',
        name = "Move Up",
        desc = "Move this up in the list of priorities.",
        order = 1,
    }
    
    local moveDown = {
        type = 'execute',
        name = "Move Down",
        desc = "Move this down in the list of priorities.",
        order = 2,
    }
    local printmenu = {
        type = 'execute',
        name = "Print",
        desc = "Print values for this category",
        order = 3,
        hidden = "ShowDebugMenus",
    }
    local printall = {
        type = 'execute',
        name = "Print All",
        desc = "Print all values for this category",
        order = 4,
        hidden = "ShowDebugMenus",
    }
    local update = {
        type = 'execute',
        name = "Update",
        desc = "Update all values for this category",
        order = 5,
        hidden = "ShowDebugMenus",
    }
    local submenusep = separator(6)
    
    local deleteKeys = {}
    local function updateOptionsTableElements(parent,fromOrder,menuType,desc,...)
        self:Debug("updateOptionsTableElements",parent,fromOrder,menuType,desc,...)
        
        local args = parent.args
        local lim = select('#',...)
        local avail,ignored,filtered = 0,0,0
        
        for k,v in pairs(args) do
            if v.order and v.order > fromOrder then
                deleteKeys[k] = true
            end
        end
        
        for i=2,lim,2 do
            local v = select(i-1,...)
            local status = select(i,...)
            
            assert(v ~= nil and type(v) == "string",parent.key.." got key "..type(v)..
                " - "..tostring(v).." from "..parent.priorityFunction)
            
            self:Debug(v,status)
            
            local element = args[v]
            deleteKeys[v] = nil
            
            if not element then
                element = {}
                args[v] = element
            end
            
            element.type = menuType
            if status then
                if status == "FILTERED" then
                    element.disabled = true
                    element.hidden = false
                    element.name = cGREY..v.." (filtered)"
                    element.desc = desc and "(filtered) "..string.format(desc,tostring(v))
                    filtered = filtered + 1
                else
                    element.disabled = false
                    element.hidden = false
                    element.name = v
                    element.desc = desc and string.format(desc,tostring(v))
                    avail = avail + 1
                end
            else
                element.disabled = true
                element.hidden = parent.childHidden
                element.name = cGREY..v
                ignored = ignored + 1
                element.desc = desc and "(unavailable) "..string.format(desc,tostring(v))
            end
            element.order = fromOrder + i
        end
        
        for k,v in pairs(deleteKeys) do
            args[k] = nil
            deleteKeys[k] = nil
        end
        
        return avail, ignored, filtered
    end
    
    local function setupProfileSubmenu(parent,modifiable,...)
        assert(parent.priorityFunction,parent.key.." priorityFunction is nil")
        local success,result = pcall(self[parent.priorityFunction],self,true)
        assert(success,"can't call self:"..parent.priorityFunction.." for "..parent.key.." - "..tostring(result))
        
        local function OnUpdate(handler,...)
            self:Debug("OnUpdate",parent.key)
            
            if not handler then handler = self end
            
            handler:Debug("OnUpdate",parent.priorityFunction,...)
            local zonesAvail, zonesIgnored, zonesFiltered = updateOptionsTableElements(
                parent,7,modifiable and 'toggle' or 'header',
                "Enable/disable LFG priority for %s",
                handler[parent.priorityFunction](handler,true,true))
            
            local k = parent.key
            local num = handler:GetPriorityNumber(k)
            local ignoreNum = handler:GetPriorityNumber(self.PRIORITY_IGNORE_HEADER)
            
            handler:UpdatePriorityCount(parent,k,v,0,0,zonesIgnored,
                zonesAvail,zonesFiltered,num >= ignoreNum)
            
            handler:UpdateAutoLFG()
        end
        
        parent.updateFunction = OnUpdate
        
        parent.childHidden = function()
            return (parent.activeCount or 0) + (parent.passiveCount or 0)
                + (parent.filteredCount or 0) > 25
        end
        
        self.OnObjectivesUpdate[parent.key] = OnUpdate
        self.OnFilterUpdate[parent.key] = OnUpdate
        
        for i=1,select('#',...) do
            local event = select(i,...)
            local onevent = "On"..event
            
            -- TODO need a better way to do this
            if not self[onevent] then
                self[onevent] = {}
                
                if not self[event] then
                    self[event] = function(handler)
                        handler:Debug(event,"event handler")
                        
                        for k,v in pairs(handler[onevent]) do
                            handler:Debug("CALLING OnUpdate for",k)
                            local ret, err = pcall(v,handler)
                            if not ret then
                                error(k.." - "..err)
                            end
                        end
                    end
                end
                self:RegisterEvent(event)
            end
            
            self[onevent][parent.key] = OnUpdate
        end
    end
    
    local asterisk = string.byte("*")
    
    temp = profileMenu.args.priorities.args
    local ignoreNum = self:GetPriorityNumber(self.PRIORITY_IGNORE_HEADER)
    temp[self.PRIORITY_IGNORE_HEADER] = {
        type = 'header',
        name = self.PRIORITY_IGNORE_HEADER,
        order = ignoreNum
    }
    
    for k,v in pairs(self.PRIORITY) do
        local num = self:GetPriorityNumber(k)
        assert(type(k) == "string","bad key value ")
        local func = v[1]
        temp[k] = {
            type = 'group',
            pass = true,
            get = function(...) 
                return self:GetPriority(k,...)
            end,
            set = function(...)
                return self:SetPriority(k,...)
            end,
            func = function(...)
                return self:FuncPriority(k,...)
            end,
            key = k,
            -- TODO Check icon (is looking) ( slots / potential groups in this selection )  Selection Name (... -> checklist of objectives)
            priorityFunction = string.byte(func) == asterisk and string.sub(func,2) or func,
            activeCount = 1,
            passiveCount = 0,
            filteredCount = 0,
            args = {
                moveUp = moveUp,
                moveDown = moveDown,
                print = printmenu,
                printall = printall,
                update = update,
                sep = submenusep,
                none = prioritySubmenuNone,
                -- list of items - inserted on update
            },
        }
        
        if string.byte(func) == asterisk then
            setupProfileSubmenu(temp[k],true,select(2,unpack(v)))
        else
            setupProfileSubmenu(temp[k],false,select(2,unpack(v)))
        end
        
        self:UpdatePriorityCount(temp[k],k,v,0,0,0,1,0,num >= ignoreNum)
        -- args, key, value, users you're in queue with, total users in this group
        self:UpdatePriorityOrder(temp[k],k,v,num,ignoreNum)
    end
    temp = nil
    
    return profileMenu
end



-- =============================================
--             ACE ADDON FUNCTIONS
-- =============================================

function lfg:OnInitialize()

    -- RUN ONCE!
    local FIXVARS = "Restore your old saved variables or remove your current ones."
    
    if self.db.profile.serialversion.check ~= self.db.char.serialversioncheck then
        -- punish the user for playing with their saved variables when upgrading!
        self:error(NAME..": Account and Character saved variables are from different versions. "..FIXVARS)
    end
    
    if self.db.profile.serialversion.check ~= bitsCount then
        -- TODO future versions will upgrade the bits
        self:error(NAME..": Modifying this addon's saved variables is not recommended. "..FIXVARS)
    end
    
    self:OnLogin()
end

function lfg:OnEnable()
    self:Debug("OnEnable")
    
    self:Allocate()
    
    -- functions to all at various update times
    self.OnLFGUpdate = Bar_OnLFGUpdate --(self,catText,query)
    self.OnObjectiveCountUpdate = Bar_OnObjectiveCountUpdate --(self,catText,query)
    self.OnCategoryCountUpdate = Bar_OnCategoryCountUpdate --(self,category)
    self.OnCategoryUpdate = Bar_OnCategoryUpdate --(self,self.categories)
    
    -- if types update while we are disabled we should rescan them
    self:ScheduleEvent("UPDATE_LFG_TYPES","UPDATE_LFG_TYPES",10,NAME)
    
    local t = self:GetLoginPause() * 60
    if t > 0 then
        self:ScheduleEvent(self.UpdateLoginState,t,self)
    end
    
    -- =============================================
    --             MENU CREATION
    -- =============================================
    
    local ACTION_ORDER = 1
    local BROWSE_ORDER = 100
    local CONFIG_ORDER = 200
    
    local browseMenu = {
        type = 'group',
        name = "Browse LFG...",
        desc = "View available players.",
        order = BROWSE_ORDER,
        args = {},
    }
    self.browse = browseMenu
    self.OnObjectivesUpdate["Bar_OnObjectivesUpdate"] = Bar_OnObjectivesUpdate
    
    local filterMenu = {
        type = 'group',
        name = "LFG Filter...",
        desc = "Checked items are permitted.",
        order = CONFIG_ORDER + 1,
        args = {
            zone = populateBitsSubmenu({
                name = "Zone Difficulty",
                desc = "Permit these zone difficulties",
                order = 1,
            },colorBits,self.GetZoneColorFilter,self.SetZoneColorFilter,self,lfg:GetColors()),
            instance = populateBitsSubmenu({
                name = "Instance Difficulty",
                desc = "Permit these instance difficulties",
                order = 2,
            },colorBits,self.GetInstanceColorFilter,self.SetInstanceColorFilter,self,lfg:GetColors()),
            quest = populateBitsSubmenu({
                name = "Quest Difficulty",
                desc = "Permit these quest difficulties",
                order = 3,
            },colorBits,self.GetQuestColorFilter,self.SetQuestColorFilter,self,lfg:GetColors()),
            size = populateBitsSubmenu({
                name = "Group Size",
                desc = "Permit these group sizes",
                order = 4,
            },sizeBits,self.GetInstanceSizeFilter,self.SetInstanceSizeFilter,self,lfg:GetInstanceSizes()),
            continent = populateBitsSubmenu({
                name = "Continent",
                desc = "Permit zones and instances in this continent",
                order = 5,
            },continentBits,self.GetContinentFilter,self.SetContinentFilter,self,lfg:GetContinents()),
        },
    }
    
    local profileMenu = {
        type = 'group',
        name = "LFG Priorities...",
        desc = "Attempt to join LFG based on these priorities.",
        order = order,
        args = {
            load = {
                type = 'group',
                pass = true,
                func = "LoadLFGProfile",
                name = "Load LFG Profile",
                desc = "Load a different generalized set of LFG priorities.",
                order = CONFIG_ORDER,
                -- args added in by ConstructProfileMenu
            },
            filter = filterMenu,
            priorities = {
                type = 'group',
                name = "Edit LFG Profile...",
                desc = "Attempting to join LFG based on these priorities.",
                order = CONFIG_ORDER + 2,
                -- args added in by ConstructProfileMenu,
            },
        },
    }

    self:ConstructProfileMenu(profileMenu,CONFIG_ORDER + 2)
    
    assert(profileMenu.args.priorities.args)
    self.priorities = profileMenu.args.priorities.args

    local status = {
        type = 'text',
        name = "Set LFG Status...",
        desc = "Change your LFG status",
        set = "SetStatus",
        get = "GetStatus",
        order = ACTION_ORDER + 1,
        validate = {
            DND = cTRANS.."1"..cRESET.."DND",
            AVAILABLE = cTRANS.."2"..cRESET.."OK",
            JOIN = cTRANS.."3"..cRESET.."Join",
            FORM = cTRANS.."4"..cRESET.."Form",
        }
    }
    local list = {
        type = 'execute',
        name = "List",
        desc = "List available LFG categories and objectives",
        func = "PrintAvailable",
        order = ACTION_ORDER + 1,
    }
    
    local scan = {
        type = 'execute',
        name = "Scan",
        desc = "Scan all LFG categories and objectives",
        func = "LFGWho",
        order = ACTION_ORDER,
    }
    
    local afklfg = {
        type = 'toggle',
        name = "Away LFG",
        desc = "Keep yourself available in the LFG system when <AFK>",
        get = "IsIgnoreAFK",
        set = "ToggleIgnoreAFK",
        order = ACTION_ORDER + 3,
    }
    
    local dndlfg = {
        type = 'toggle',
        name = "DND LFG",
        desc = "Keep yourself available in the LFG system when <DND>",
        get = "IsIgnoreDND",
        set = "ToggleIgnoreDND",
        order = ACTION_ORDER + 4,
    }
    
    local taxilfg = {
        type = 'toggle',
        name = "LFG while on Flightpath",
        desc = "Keep yourself available to proximity-dependant zones when on a flightpath",
        get = "IsIgnoreTaxi",
        set = "ToggleIgnoreTaxi",
        order = ACTION_ORDER + 5,
    }
    
    local grouplfg = {
        type = 'toggle',
        name = "LFG while in a group",
        desc = "Continue to be available to the LFG system when in a group",
        get = "IsAvailableInGroup",
        set = "ToggleAvailableInGroup",
        order = ACTION_ORDER + 6,
    }
    
    local loginpause = {
        type = 'range',
        name = "Login Grace Period",
        desc = "Number of minutes to wait after login to start to LFG",
        get = "GetLoginPause",
        set = "SetLoginPause",
        min = 0,
        max = 30,
        step = 1,
        order = ACTION_ORDER + 7,
    }
    
    local autoform = {
        type = 'toggle',
        name = "Auto-Form Group",
        desc = "Try to form a new group after trying to join a group without success",
        get = "IsAutoForm",
        set = "ToggleAutoForm",
        order = ACTION_ORDER + 8,
    }
    
    local autoformpause = {
        type = 'range',
        name = "Auto-Form Grace Period",
        desc = "Number of minutes to wait after trying to join a group to try to form a new group",
        get = "GetAutoFormPause",
        set = "SetAutoFormPause",
        min = 0,
        max = 30,
        step = 1,
        order = ACTION_ORDER + 9,
    }
    
    -- key = {
    --      type = 'group'
    --      pass = true
    --      func = function(key) return self:PriorityExecute(name, key) end
    --      set = function(key,value) return self:PrioritySet(name, key, value) end
    --      get = function(key) return self:PriorityGet(name, key) end
    --      -- call a function that returns this priority's SetLookingForGroup parameters
    --      getLFGTarget = "GetLFGTargetFunction"
    --      -- list of events this priority should register for - when they occur this 
    --      updateEvents = { list of AceEvent keys to bind to }
    --      args = prioritySubmenu
    -- }
    
    -- options table elements, order == preference
    --  filter: zone color - Grey,Green,Yellow,Orange,Red
    --  filter: instance color - "
    --  filter: instance size - 5,10,15,20,25,40
    --  filter: continent - Kalimdor, Eastern Kingdoms, Outland
    -- Current Zone
    -- Adjacent Zones
    -- All Zones
    -- Current Zone's Instances
    -- Adjacent Zones' Instances
    -- All Instances
    -- Battleground
    
    -- FUTURE:
    -- Current World PvP Zones
    -- Adjacent World PvP Zones
    -- All World PvP Zones
    
    
    -- 4 States: - click toggles between them
    --  Do No Disturb - quit all LFG's
    --  Available - join your preferred LFG profiles
    --  Join Group - join your preferred LFG profiles, set auto-join
    --  Form Group - if no group is joined (escalates from join group when noone is LFM and 2 minutes have passed)
    
    -- Display:
    -- LFG: State (#Preferred/#Available)
    
    self.OnMenuRequest = {
        type = 'group',
        handler = lfg,
        args = {
            scan = scan,
            status = status,
            config = {
                type = 'group',
                order = ACTION_ORDER + 2,
                name = "Config...",
                desc = "Set configuration options",
                args = {
                    afklfg = afklfg,
                    dndlfg = dndlfg,
                    taxilfg = taxilfg,
                    grouplfg = grouplfg,
                    loginpause = loginpause,
                    autoform = autoform,
                    autoformpause = autoformpause,
                },
            },
            
            browseSep = separator(BROWSE_ORDER - 1),
            browse = browseMenu,
            priority = profileMenu,
            
--            configSep = separator(CONFIG_ORDER - 1),
--            filter = filterMenu,
            endSep = separator(1000),
        },
    }
    
    self:RegisterChatCommand({"/grouper","/grouperlfg"},{
        type = 'group',
        args = {
            list = list,
            scan = scan,
            status = status,
            afklfg = afklfg,
            dndlfg = dndlfg,
            taxilfg = taxilfg,
            grouplfg = grouplfg,
            loginpause = loginpause,
            autoform = autoform,
            autoformpause = autoformpause,
            test = {
                type = 'group',
                name = "Test Menu",
                desc = "Test Functions",
                args = {
                    updateall = {
                        type = 'execute',
                        name = "Update All",
                        desc = "Update All",
                        func = function() self:UpdateObjectives() end
                    },
                },
            },
        },
    })
    
    self:RegisterChatCommand({"/lfgwho"},function(text) self:LFGWho(text) end)
    
    self:RegisterChatCommand({"/groupertest"},function(text) self[text](self) end)
    
    -- =============================================
    --                  EVENTS
    -- =============================================
    
    -- your lfg status/available LFG items updated
    self:RegisterEvent("LFG_UPDATE")
    
    -- recieved lfg query response?
    self:RegisterEvent("UPDATE_LFG_LIST")
    
    -- update available types
    self:RegisterEvent("UPDATE_LFG_TYPES")
    
    -- (legacy event?)
    self:RegisterEvent("MEETINGSTONE_CHANGED")
    
    -- watch for AFK/DND
    self:RegisterEvent("CHAT_MSG_SYSTEM")
    
    -- watch for Taxi (flightpath)
    self:RegisterEvent("TAXIMAP_CLOSED")
    self:RegisterEvent("PLAYER_CONTROL_GAINED")
    
    self:RegisterEvent("PARTY_MEMBERS_CHANGED")
    self:RegisterEvent("PARTY_LEADER_CHANGED")
    self:RegisterEvent("RAID_ROSTER_UPDATE")
    
    self:UpdateAFKState(nil,true)
    self:UpdateTaxiState(true)
    self:UpdateGroupState(true)
    self:UpdateAutoLFG()
    
    self:SetIcon(self:GetLFGStateIcon(STATE_VALUE[stateBits:MaskHash(self.state,"PRIMARY")]))
    
    -- =============================================
    --                  HOOKS
    -- =============================================
    
    self:SecureHook("LFGQuery","Hook_LFGQuery")
    self:SecureHook("SetLFGNameCriteria","Hook_SetLFGNameCriteria")
end

function lfg:Hook_LFGQuery(...)
    self:Debug("SecureHook: LFGQuery",...)
    local query = select(1,...)..';'..select(2,...)
    self.currentQuery = query
    self:RemoveQuery(query)
end

function lfg:Hook_SetLFGNameCriteria(frame, id, ...)
    self:Debug("SecureHook: SetLFGNameCriteria",frame, id, ...)
    local frameId = frame:GetID()
    local category = frame.selectedType
    
    -- TODO detect user set LFG's vs auto-set LFG's
end

function lfg:OnDisable()
	self.who = nil
    self.cache = nil
    self.queue = nil
    
    -- TODO remove the auto-set LFG's but not user-set ones
--    ClearLookingForGroup()
--    ClearLookingForMore()
	-- redundant! self:UnregisterAllEvents()
end


-- =============================================
--                  EVENTS
-- =============================================

function lfg:ZONE_CHANGED_NEW_AREA()
    self:ScheduleEvent("OnZoneChange",self.OnZoneChange,4,self)
end

function lfg:OnZoneChange()
    self:Debug("OnZoneChange")
    
    if not self.OnZONE_CHANGED_NEW_AREA then return end
    
    for k,v in pairs(self.OnZONE_CHANGED_NEW_AREA) do
        self:Debug("CALLING OnUpdate for",k)
        local ret, err = pcall(v,self)
        if not ret then
            error(k.." - "..err)
        end
    end
end

-- =============================================
--           ZONE INFORMATION FUNCTIONS
-- =============================================

local function CurrentZone()
    return GetRealZoneText()
end

function lfg:GetInstanceOrZoneDifficulty(objective)
    return GetColorWord(tourist:GetLevelColor(objective))
end

function lfg:GetInstanceGroupSize(objective)
    local size = tourist:GetInstanceGroupSize(objective)
    
    if size == 0 then
        self:Debug("GetInstanceGroupSize returned 0 for",objective)
    end
    
    return size
end

function lfg:GetInstanceOrZoneContinent(objective)
    return tourist:GetContinent(objective)
end

function lfg:IsZoneFiltered(zone)
    local colorFilter = self:GetZoneColorFilter()
    local zoneColor = self:GetInstanceOrZoneDifficulty(zone)
    
    if zoneColor and not colorBits:HashContains(colorFilter,zoneColor) then
        self:Debug("failed zone difficulty",zoneColor)
        return true
    end
    
    local continentFilter = self:GetContinentFilter()
    local zoneContinent = self:GetInstanceOrZoneContinent(zone)
    
    if zoneContinent and not continentBits:HashContains(continentFilter,zoneContinent) then
        self:Debug("failed continent",zoneContinent)
        return true
    end
    
    local groupSizeFilter = self:GetInstanceSizeFilter()
    if not sizeBits:HashContains(groupSizeFilter,5) then
        self:Debug("failed group size",5)
        return true
    end
    
    return false
end

--GetInstanceColorFilter
--GetInstanceSizeFilter
--GetContinentFilter 

function lfg:IsInstanceFiltered(zone,heroic)
    local colorFilter = self:GetZoneColorFilter()
    local zoneColor = heroic and "RED" or self:GetInstanceOrZoneDifficulty(zone)
    
    if zoneColor and not colorBits:HashContains(colorFilter,zoneColor) then
        self:Debug("failed zone difficulty",zoneColor)
        return true
    end
    
    local continentFilter = self:GetContinentFilter()
    local zoneContinent = self:GetInstanceOrZoneContinent(zone)
    
    if zoneContinent and not continentBits:HashContains(continentFilter,zoneContinent) then
        self:Debug("failed continent",zoneContinent)
        return true
    end
    
    local groupSizeFilter = self:GetInstanceSizeFilter()
    local groupSize = self:GetInstanceGroupSize(zone)
    
    if groupSize and not sizeBits:HashContains(groupSizeFilter,groupSize) then
        self:Debug("failed group size",5)
        return true
    end
    
    return false
end

--GetInstanceColorFilter
--GetInstanceSizeFilter

function lfg:IsBattlegroundFiltered(bg)
    return self:IsInstanceFiltered(bg)
end

--GetQuestColorFilter
--GetInstanceSizeFilter

function lfg:IsQuestFiltered(quest)
end

local function nlist(present, filtered, returnAll, showFiltered)
--    if returnAll then
--        return true
--    elseif present then
--        if showFiltered or not filtered then
--            return true
--        end
--    end
    
    return returnAll or present and (showFiltered or not filtered)
end

local function ndis(present, filtered, returnAll, showFiltered)
    if present then
        if filtered then
            return "FILTERED"
        end
        return "OK"
    end
    return false
end

-- return zone, true/false status
function lfg:GetCurrentZone(returnAll,showfiltered)
    local z = CurrentZone()
    
    if string.len(z) > 0 then
        local present = self.zones[z] ~= nil or self.instances[z] ~= nil
            or self.battlegrounds[z] ~= nil
        local filtered = self:IsZoneFiltered(z)
        
        if nlist(present, filtered, returnAll, showfiltered) then
            return z, ndis(present, filtered, returnAll, showfiltered)
        end
    end
end

local clearZoneCache, addZoneCache, returnZoneCache
do
    local zoneCache = {}
    function clearZoneCache()
        while #zoneCache > 0 do
            table.remove(zoneCache)
        end
    end
    
    if SAFE_MODE then
        function addZoneCache(key, value)
            if #zoneCache > 250 then error("zoneCache sanity check (100) failed",2) end
            table.insert(zoneCache,key)
            table.insert(zoneCache,value)
        end
    else
        function addZoneCache(key, value)
            table.insert(zoneCache,key)
            table.insert(zoneCache,value)
        end
    end
    
    function returnZoneCache()
        return unpack(zoneCache)
    end
end

function lfg:GetAdjacentZones(returnAll,showfiltered)
    clearZoneCache()
    
    local current, present, filtered = CurrentZone()
    for z in tourist:IterateBorderZones(current) do
        if tourist:IsZone(z) then
            present = self.zones[z] ~= nil
            filtered = self:IsZoneFiltered(z)
            if nlist(present, filtered, returnAll, showfiltered) then
                addZoneCache(z, ndis(present, filtered, returnAll, showfiltered))
            else
                self:Debug("not showing",z,
                    present and "present" or "not present",
                    filtered and "filtered" or "not filtered",
                    returnAll and "return all" or "",
                    showfiltered and "show filtered" or "")
            end
        elseif not tourist:IsZoneOrInstance(z) then
            -- TODO option to transverse transports
            for y in tourist:IterateBorderZones(z) do
                if y ~= current and tourist:IsZone(y) then
                    present = self.zones[y] ~= nil
                    filtered = self:IsZoneFiltered(y)
                    if nlist(present, filtered, returnAll, showfiltered) then
                        addZoneCache(y, ndis(present, filtered, returnAll, showfiltered))
                    else
                        self:Debug("not showing",y,
                            present and "present" or "not present",
                            filtered and "filtered" or "not filtered",
                            returnAll and "return all" or "",
                            showfiltered and "show filtered" or "")
                    end
                else
                    self:Debug("not zone or current",y)
                end
            end
        else
            self:Debug("not zone",z)
        end
    end
    
    return returnZoneCache()
end

function lfg:GetAllZones(returnAll,showfiltered)
    clearZoneCache()
    
    local present, filtered
    for z in tourist:IterateZones() do
        if tourist:IsZone(z) then
            present = self.zones[z]
            filtered = self:IsZoneFiltered(z)
            if nlist(present,filtered,returnAll,showfiltered) then
                addZoneCache(z,ndis(present, filtered, returnAll, showfiltered))
            end
        end
    end
    
    return returnZoneCache()
end

function lfg:GetCurrentZonesInstances(returnAll,showfiltered)
    clearZoneCache()
    
    local present, filtered
    for z in tourist:IterateZoneInstances(CurrentZone()) do
        present = self.instances[z]
        filtered = self:IsInstanceFiltered(z)
        if nlist(present,filtered,returnAll,showfiltered) then
            addZoneCache(z, ndis(present, filtered, returnAll, showfiltered))
        end
    end
    
--    if returnAll then
--        for z in tourist:IterateZoneInstances(CurrentZone()) do
--            addZoneCache(z,self.instances[z] ~= nil)
--        end
--    else
--        for z in tourist:IterateZoneInstances(CurrentZone()) do
--            if self.instances[z] then
--                addZoneCache(z,true)
--            end
--        end
--    end
    
    return returnZoneCache()
end

function lfg:GetAdjacentZonesInstances(returnAll,showfiltered)
    clearZoneCache()
    
    local current,present,filtered = CurrentZone()
    for z in tourist:IterateBorderZones(current) do
        if tourist:IsZone(z) then
            for i in tourist:IterateZoneInstances(z) do
                present = self.instances[i]
                filtered = self:IsInstanceFiltered(i)
                if nlist(present,filtered,returnAll,showfiltered) then
                    addZoneCache(i, ndis(present, filtered, returnAll, showfiltered))
                end
            end
        elseif not tourist:IsZoneOrInstance(z) then
            -- TODO option to transverse transports
            for y in tourist:IterateBorderZones(z) do
                if y ~= current and tourist:IsZone(y) then
                    for i in tourist:IterateZoneInstances(y) do
                        present = self.instances[i]
                        filtered = self:IsInstanceFiltered(i)
                        if nlist(present,filtered,returnAll,showfiltered) then
                            addZoneCache(i, ndis(present, filtered, returnAll, showfiltered))
                        end
                    end
                end
            end
        end
    end
    
    return returnZoneCache()
end

function lfg:GetAllInstances(returnAll,showfiltered)
    clearZoneCache()
    
    -- TODO can't optimize this yet...
    -- iterating self.instances shows the disparity between
    -- zone names and LFG returned instance names
    if returnAll then
        for z in tourist:IterateInstances() do
            present = self.instances[z]
            filtered = self:IsInstanceFiltered(z)
            if nlist(present,filtered,returnAll,showfiltered) then
                addZoneCache(z,ndis(present, filtered, returnAll, showfiltered))
            end
        end
    else
        for k,v in pairs(self.instances) do
            present = true
            filtered = self:IsInstanceFiltered(k)
            if nlist(present,filtered,returnAll,showfiltered) then
                addZoneCache(k,ndis(present, filtered, returnAll, showfiltered))
            end
        end
    end
    
    return returnZoneCache()
end

function lfg:GetAllBattlegrounds(returnAll,showfiltered)
    clearZoneCache()
    
    local present,filtered
    for z in tourist:IterateBattlegrounds() do
        present = self.battlegrounds[z]
        filtered = self:IsBattlegroundFiltered(z)
        if nlist(present, filtered, returnAll, showfiltered) then
            addZoneCache(z, ndis(present, filtered, returnAll, showfiltered))
        end
    end
    
    return returnZoneCache()
end

function lfg:GetAdjacentWorldPvPZones(returnAll,showfiltered)
    clearZoneCache()
    
    local current, present, filtered = CurrentZone()
    for z in tourist:IterateBorderZones(current) do
        if tourist:IsPvPZone(z) then
            present = self.zones[z]
            filtered = self:IsZoneFiltered(z)
            if nlist(present,filtered,returnAll,showfiltered) then
                addZoneCache(z,ndis(present, filtered, returnAll, showfiltered))
            end
        elseif not tourist:IsZoneOrInstance(z) then
            -- TODO option to transverse transports
            for y in tourist:IterateBorderZones(z) do
                if y ~= current and tourist:IsPvPZone(y) then
                    present = self.zones[y]
                    filtered = self:IsZoneFiltered(y)
                    if nlist(present,filtered,returnAll,showfiltered) then
                        addZoneCache(y,ndis(present, filtered, returnAll, showfiltered))
                    end
                end
            end
        end
    end
    
    return returnZoneCache()
end

function lfg:GetAllPvPZones(returnAll,showfiltered)
    clearZoneCache()
    
    local present
    for z in tourist:IteratePvPZones() do
        present = self.zones[z]
        filtered = self:IsZoneFiltered(z)
        if nlist(present,filtered,returnAll,showfiltered) then
            addZoneCache(z,ndis(present, filtered, returnAll, showfiltered))
        end
    end
    
    return returnZoneCache()
end

function lfg:GetInstanceZone(instance)
    if tourist:IsInstance(instance) then
        return instance
    end
    
    string.match(instance,"(.-[^ -]) ?\- ?([^ ].*)")
end

-- =============================================
--           QUEST INFORMATION FUNCTIONS
-- =============================================

function lfg:GetQuestDifficulty(category,objective)
    local name, level, tag, objective, isComplete, nObjectives, zone, id = quixote:GetQuest(objective)
    
    return self:GetLevelColor(level)
end

function lfg:GetQuestGroupSize(category,objective)
    local name, level, tag, objective, isComplete, nObjectives, zone, id = quixote:GetQuest(objective)
    
    if tag == "Raid" then
        return "5+"
    else
        return 5
    end
end

function lfg:GetCurrentZonesQuests(returnAll,showfiltered)
    clearZoneCache()
    
    local present, title, level, tag, objective, complete, nlb, zone
    for i in quixote:QuestsByZone(CurrentZone()) do
        title, level, tag, objective, complete,
            nlb, zone = quixote:GetQuestById(i)
        present = title and self.quests[title] and true or false
        if title and not complete and (returnAll or present) then
            addZoneCache(title,present and not filtered)
        end
    end
    
    return returnZoneCache()
end

function lfg:GetAdjacentZonesQuests(returnAll,showfiltered)
    clearZoneCache()
    
    local current = CurrentZone()
    local present, title, level, tag, objective, complete, nlb, zone
    for z in tourist:IterateBorderZones(current) do
        if tourist:IsZoneOrInstance(z) then
            for i in quixote:QuestsByZone(z) do
                title, level, tag, objective, complete,
                    nlb, zone = quixote:GetQuestById(i)
                present = title and self.quests[title] and true or false
                if title and not complete and (returnAll or present) then
                    addZoneCache(title,present and not filtered)
                end
            end
        else
            -- TODO option to transverse transports
            for y in tourist:IterateBorderZones(z) do
                if y ~= current and tourist:IsZoneOrInstance(y) then
                    for i in quixote:QuestsByZone(y) do
                        title, level, tag, objective, complete,
                            nlb, zone = quixote:GetQuestById(i)
                        present = title and self.quests[title] and true or false
                        if title and not complete and (returnAll or present) then
                            addZoneCache(title,present and not filtered)
                        end
                    end
                end
            end
        end
    end
    
    return returnZoneCache()
end

function lfg:GetAllQuests(returnAll,showfiltered)
    clearZoneCache()
    
    local present
    for id, title, level, tag, objective, complete, nlb, zone in quixote:IterateQuests() do
        present = title and self.quests[title]
        
        if nlist(present, filtered, returnAll, showfiltered) then
            addZoneCache(title, ndis(present, filtered, returnAll, showfiltered))
        end
        
--        if title and not complete and (returnAll or present) then
--            addZoneCache(title,present and not filtered)
--        end
    end
    
    return returnZoneCache()
end

