local lfg = {}
getfenv(0)["Grouper-1.0"] = lfg

--local tablet = AceLibrary and AceLibrary("Tablet-2.0") or error("Library Tablet-2.0 is required") -- when I get around to tooltips
local Bits = AceLibrary and AceLibrary("Bits-1.0") or error("Library Bits-1.0 is required")
--local Dewdrop = AceLibrary and AceLibrary("Dewdrop-2.0") -- TODO check out Dewdrop and update fubar menu

-- =============================================
--             UTILITY FUNCTIONS
-- =============================================

local NAME = "Grouper"
local Sys_GetLFGTypes, Sys_GetLFGTypeEntries = GetLFGTypes, GetLFGTypeEntries

local mathmin, mathmax = math.min, math.max
local stringfind, stringsub = string.find, string.sub
local _,temp

local function packInto(list,...)
    for i=1,select('#',...) do
        list[i] = select(i,...)
    end
end

local function Split(query,delim)
    if not query then return end
    local x,y = stringfind(query,delim,1,true)
    if not x then return query end
    return stringsub(query,1,x-1),Split(stringsub(query,y+1),delim)
end

function lfg:Allocate()
    self:Print("THIS IS AN ALPHA RELEASE. Please see http://www.wowace.com/wiki/Grouper",
        "for current features and known issues.")
    
    -- =============================================
    --             RUNTIME VARIABLES
    -- =============================================
    
    self.categories = {}
    -- [number] = category
    self.objectives = {}
    -- [query] = objective text
    
    self.criteria = {}
    -- [criteria] = query
    
    self.zones = {}
    self.instances = {}
    self.heroic = {}
    self.quests = {}
    self.battlegrounds = {}
    
    self.playerLevel = UnitLevel("player")
    
	self.who = {}
	-- [name] =
	-- 	time;level;zone;class;criteria1;criteria2;criteria3;comment;numparty;isLFM
    
    self.whoCount = 0
    -- needed because who is a map, no indexes
    
    self.partyWho = {}
    -- [name] =
    --  name,level,class;name,level,class;...
    
	self.cache = {}
	-- [query] = time, list of names
    
    self.categoryCache = {}
    -- [number] = time, objectives scanned, total objectives, list of names
	
	self.queue = {}
	-- processing queue table of category;objective = time
	
    -- functions to call when an objectives update completes (self)
    self.OnObjectivesUpdate = {}
    
    -- funcs to call when preference filters are changed
    self.OnFilterUpdate = {}
end
-- =============================================
--                 CONSTANTS
-- =============================================

function lfg:GetColors()
    return "RED","ORANGE","YELLOW","GREEN","GREY"
end

function lfg:GetInstanceSizes()
    return 5,10,15,20,25,40
end

function lfg:GetContinents()
    return "Eastern Kingdoms", "Kalimdor", "Outland"
end

-- [name] = { update func, update event, update event... }
lfg.PRIORITY = {
    ["Current Zone"] = { "GetCurrentZone", "ZONE_CHANGED_NEW_AREA", "PLAYER_LEVEL_UP" },
    ["Adjacent Zones"] = { "GetAdjacentZones", "ZONE_CHANGED_NEW_AREA", "PLAYER_LEVEL_UP" }, -- req Tourist
    ["All Zones"] = { "GetAllZones", "PLAYER_LEVEL_UP" },
    ["Select Zone..."] = { "*GetAllZones", "PLAYER_LEVEL_UP" },
--    ["Closest Instance"] = { "GetClosestInstance" }, -- FUTURE needs change to TouristLib
    ["Current Zone's Instances"] = { "GetCurrentZonesInstances", "ZONE_CHANGED_NEW_AREA", "PLAYER_LEVEL_UP" }, -- req Tourist
    ["Adjacent Zones' Instances"] = { "GetAdjacentZonesInstances", "ZONE_CHANGED_NEW_AREA", "PLAYER_LEVEL_UP" }, -- req Tourist
    -- TODO Heroic Dungeons
    ["All Instances"] = { "GetAllInstances", "PLAYER_LEVEL_UP" },
    ["Select Instance..."] = { "*GetAllInstances", "PLAYER_LEVEL_UP" },
    ["All Battlegrounds"] = { "GetAllBattlegrounds", "PLAYER_LEVEL_UP" }, -- FUTURE needs Blizzard LFG update
    ["Select Battleground..."] = { "*GetAllBattlegrounds", "PLAYER_LEVEL_UP" }, -- FUTURE needs Blizzard LFG update
--    ["Closest World PvP Zone"] = { "GetClosestWorldPvPZone" }, -- req Tourist
    ["Adjacent World PvP Zones"] = { "GetAdjacentWorldPvPZones", "ZONE_CHANGED_NEW_AREA", "PLAYER_LEVEL_UP" }, -- req Tourist
    ["All World PvP Zones"] = { "GetAllPvPZones", "PLAYER_LEVEL_UP" }, -- req Tourist
    ["Select World PvP Zone..."] = { "*GetAllPvPZones", "PLAYER_LEVEL_UP" }, -- req Tourist
    -- req Quixote
    ["Current Zone's Quests"] = { "GetCurrentZonesQuests", "Quixote_Update", "PLAYER_LEVEL_UP" },
    ["Adjacent Zones' Quests"] = { "GetAdjacentZonesQuests", "Quixote_Update", "PLAYER_LEVEL_UP" },
    ["All Quests"] = { "GetAllQuests", "Quixote_Update", "PLAYER_LEVEL_UP" },
    ["Select Quest..."] = { "*GetAllQuests", "Quixote_Update", "PLAYER_LEVEL_UP" },
}
lfg.PRIORITY_IGNORE_HEADER = "----- Not Joining -----"

lfg.profiles = {}

-- Check icon (is looking) ( slots / potential groups in this selection )  Selection Name (... -> checklist of objectives)

do
    local profiles = {
        {
            "PvP",
            {
                "Select Battleground...",
                "Select World PvP Zone...",
                "Select Zone...",
                "Closest World PvP Zone",
                "Adjacent World PvP Zones",
                "Current Zone",
                "All World PvP Zones",
                "All Battlegrounds",
                "Adjacent Zones",
                "All Zones",
                lfg.PRIORITY_IGNORE_HEADER,
                "Select Quest...",
                "Select Instance...",
                "Current Zone's Quests",
                "Closest Instance",
                "Current Zone's Instances",
                "Adjacent Zones' Instances",
                "Adjacent Zones' Quests",
                "All Instances",
                "All Quests",
            },
            "Looking for PvP action",
        },
        {
            "Questing",
            {
                "Select Quest...",
                "Select Zone...",
                "Select Instance...",
                "Current Zone's Quests",
                "Adjacent Zones' Quests",
                "Current Zone",
                "All Quests",
                "Adjacent Zones",
                "All Zones",
                lfg.PRIORITY_IGNORE_HEADER,
                "Select Battleground...",
                "Select World PvP Zone...",
                "Closest Instance",
                "Current Zone's Instances",
                "Adjacent Zones' Instances",
                "All Instances",
                "All Battlegrounds",
                "Closest World PvP Zone",
                "Adjacent World PvP Zones",
                "All World PvP Zones",
            },
            "Need to finish some quests",
        },
        {
            "Current Zone",
            {
                "Select Zone...",
                "Current Zone",
                "Adjacent Zones",
                "Current Zone's Quests",
                "Adjacent Zones' Quests",
                "All Zones",
                lfg.PRIORITY_IGNORE_HEADER,
                "Closest World PvP Zone",
                "Adjacent World PvP Zones",
                "All World PvP Zones",
                "Select World PvP Zone...",
                "Closest Instance",
                "Current Zone's Instances",
                "Adjacent Zones' Instances",
                "All Instances",
                "Select Instance...",
                "All Battlegrounds",
                "Select Battleground...",
                "All Quests",
                "Select Quest...",
            },
            "Looking for quick & nearby group",
        },
        {
            "Instancing",
            {
                "Select Instance...",
                "Closest Instance",
                "Current Zone's Instances",
                "Adjacent Zones' Instances",
                "All Instances",
                lfg.PRIORITY_IGNORE_HEADER,
                "Current Zone",
                "Adjacent Zones",
                "All Zones",
                "Select Zone...",
                "All Battlegrounds",
                "Select Battleground...",
                "Closest World PvP Zone",
                "Adjacent World PvP Zones",
                "All World PvP Zones",
                "Select World PvP Zone...",
                "Current Zone's Quests",
                "Adjacent Zones' Quests",
                "All Quests",
                "Select Quest...",
            },
            "Looking for an instance run",
        },
        {
            "Default",
            {
                "Current Zone",
                lfg.PRIORITY_IGNORE_HEADER,
                "Adjacent Zones",
                "All Zones",
                "Select Zone...",
                "Closest Instance",
                "Current Zone's Instances",
                "Adjacent Zones' Instances",
                "All Instances",
                "Select Instance...",
                "All Battlegrounds",
                "Select Battleground...",
                "Closest World PvP Zone",
                "Adjacent World PvP Zones",
                "All World PvP Zones",
                "Select World PvP Zone...",
                "Current Zone's Quests",
                "Adjacent Zones' Quests",
                "All Quests",
                "Select Quest...",
            },
            "Auto-LFG by GrouperLFG Addon",
        },
    }
    
    -- experimental trading off 2 concatenations per profile load (rare) vs 5 x 64 bytes
    -- returns i, key,ORDER table,NOTE
    function lfg:GetLFGProfile(key)
        for i, k, order, note in self:IterateLFGProfiles() do
            if k == key then
                return i, k, order, note
            end
        end
    end
    
    local function itr(_,k)
        local v
        k,v = next(profiles,k)
        
        if not k then return end
        
        return k, unpack(v)
    end
    
    -- returns i, key,ORDER table,NOTE
    function lfg:IterateLFGProfiles()
        return itr,nil,nil
    end
end

do
    local state = {
        DND = {
            "AVAILABLE",            -- next state
            "DND",                  -- short text
            "Do No Disturb",        -- text
            "Icon\\blue box.tga",   -- icon
        },
        AVAILABLE = {
            "JOIN",
            "OK",
            "Available",
            "Icon\\blue check.tga",
        },
        JOIN = {
            "DND",
            "Join",
            "Join Group",
            "Icon\\green check.tga",
        },
        FORM = {
            "AVAILABLE",
            "Form",
            "Form Group",
            "Icon\\orange x.tga",
        },
    }
    
    -- key, next, short, text, icon
    function lfg:GetLFGStateInfo(key)
        if state[key] then
            return key, unpack(state[key])
        end
    end
    
    function lfg:GetLFGStateNext(key)
        if state[key] then
            return state[key][1]
        end
    end
    
    function lfg:GetLFGStateShort(key)
        if state[key] then
            return state[key][2]
        end
    end
    
    function lfg:GetLFGStateText(key)
        if state[key] then
            return state[key][3]
        end
    end
    
    function lfg:GetLFGStateIcon(key)
        if state[key] then
            return state[key][4]
        end
    end
    
    local function itr(_,k)
        return next(state,k)
    end

    function lfg:IterateLFGStateInfo()
        return itr,nil,nil
    end
end


local stateBits = Bits:new(
    -- primary state
    "DND",          -- player selected DND
    "AVAILABLE",    -- player selected available
    "JOIN",         -- player selected join
    "FORM",         -- player selected form
    
    "AUTO",         -- auto upgrade/downgrade of state
    "TAXI",         -- player on a taxi, will transition to primary state
    "GROUPED",      -- player is grouped, will transition to available or dnd
    "GROUPEDOK",    -- modifies grouped to permit available
    "LEADER",       -- player is group leader
    "AFK"           -- player is <AFK> or <DND> and preferences recognize these
)

-- player logged in, will transition to available or dnd per preferences
stateBits:Alias("LOGIN","AUTO","DND")
-- join didn't find a group, attempting to form, (if joining a group transitions to available grouped)
stateBits:Alias("FORM AUTO","AUTO","FORM")
stateBits:Alias("PRIMARY","DND","AVAILABLE","JOIN","FORM")

local STATE_VALUE = stateBits:ReverseMap()

local DND = stateBits:Get("DND")
-- compatible with all modifiers
local AVAILABLE = stateBits:Get("AVAILABLE")
-- suppressed by GROUPED
-- suppressed by AFK
-- partially suppressed by TAXI
local JOIN = stateBits:Get("JOIN")
-- GROUPEDOK transitions to AVAILABLE
-- LEADER transitions to FORM
-- partially suppressed by TAXI
-- suppressed by AFK
local FORM = stateBits:Get("FORM")
-- GROUPEDOK transitions to AVAILABLE
-- suppressed by AFK

local STATE_TOGGLE = {
}

for k,v in pairs({DND = "AVAILABLE", AVAILABLE = "JOIN", JOIN = "DND", FORM = "AVAILABLE"}) do
    STATE_TOGGLE[stateBits:Get(k)] = stateBits:Get(v)
end

local UPDATE_ICON = "Icon\\orange box.tga"

local asterisk = string.byte("*")

local function LinkPlayer(player,text)
    return "|Hplayer:"..player.."|h"..text.."|h"
end

-- =============================================
--         DECLARING LOCAL FUNCTIONS
-- =============================================

local newGetBits, newSetBits, AddToCache, RemoveFromCache, GetRemovedCategoryAndCriteria

function lfg:GetPriorityNumber(lfgAction)
    local pri = self:GetPriorities()
    assert(lfgAction)
    assert(pri)
    for i,v in ipairs(pri) do
        if lfgAction == v then
            return i
        end
    end
    
    -- TODO this shouldn't happen
--    error("No such LFG action '"..lfgAction.."' in self:GetPriorities()")
    table.insert(pri,lfgAction) -- HACK
end

function lfg:ChangePriorityNumber(mod,lfgAction,lfgNumber)
    self:Debug("ChangePriorityNumber",mod,lfgAction,lfgNumber)
    
    if not lfgNumber then
        lfgNumber = self:GetPriorityNumber(lfgAction)
    end
    
    local pri = self:GetPriorities()
    
    if not lfgAction then
        lfgAction = pri[lfgNumber]
    end
    
    local count = #pri
    
    local t = table.remove(pri,lfgNumber)
    if lfgAction then
        assert(t == lfgAction)
    end
    
    local newNumber = lfgNumber - mod
    if newNumber < 1 then
        newNumber = 1
    end
    if newNumber > count then
        newNumber = count
    end
    
    t = pri[newNumber]
    
    table.insert(pri,newNumber, lfgAction)
    
    
--    self.OnObjectivesUpdate[lfgAction](self)
--    self.OnObjectivesUpdate[t](self)
    
    self:UpdatePriorityOrder(nil,lfgAction)
    self:UpdatePriorityOrder(nil,t)
end

function lfg:SetStatus(key,button)
    self:Debug("SetStatus",key,button)
    local state = stateBits:RemoveFromHash(self.state,"PRIMARY","AUTO")
    self.state = stateBits:AddToHash(state,key)
    
    self:SetInitialState(key)
    
    local t = self:GetAutoFormPause() * 60
    if key == "JOIN" and t > 0 then
        self:Debug("Scheduling UpdateAutoFormState")
        self:ScheduleEvent(self.UpdateAutoFormState,t,self)
    end
    
    self:UpdateText()
    self:UpdateAutoLFG()
    if button then
        self:ProcessQuery(true)
    end
end

function lfg:GetStatus()
    local pri = stateBits:MaskHash(self.state,"PRIMARY")
    self:Debug("GetStatus",STATE_VALUE[pri])
    return STATE_VALUE[pri]
end

function lfg:ToggleBusyIcon(userClick)
    self:Debug("ToggleBusyIcon")
    
    local now = time()
    
    if not userClick and self.nextIconUpdate and self.nextIconUpdate > now then return end
    
    if not self.updateIcon or self.updateIcon == 0 then
        self.updateIcon = 1
        self:SetIcon(UPDATE_ICON)
    else
        self.updateIcon = 0
        self:SetIcon(self:GetLFGStateIcon(STATE_VALUE[stateBits:MaskHash(self.state,"PRIMARY")]))
    end
    
    self.nextIconUpdate = now + .5
end

function lfg:IsBusyIcon()
    return self.updateIcon
end

function lfg:IsShowingBusyIcon()
    return self.updateIcon == 1
end

function lfg:StartBusyIcon()
    self:ToggleBusyIcon()
    self:ScheduleRepeatingEvent(NAME.."ICON_FLASH",self.ToggleBusyIcon,2,self)
end

function lfg:CancelBusyIcon()
    self:Debug("Canceling icon flash")
    self.updateIcon = nil
    self:CancelScheduledEvent(NAME.."ICON_FLASH")
end

function lfg:OnClick(button)
    self:ProcessQuery(true)
    
    if self:IsBusyIcon() then
        -- TODO is the flashing from clicks too erratic?
        self:ToggleBusyIcon(true)
    end
    
    if IsControlKeyDown() then
        local state = stateBits:RemoveFromHash(self.state,"PRIMARY","AUTO")
        self.state = stateBits:AddToHash(state,"FORM")
        self:SetInitialState("JOIN")
        
        self:UpdateText()
        self:UpdateAutoLFG()
    elseif IsShiftKeyDown() then
        local pri,opt = stateBits:MaskHash(self.state,"PRIMARY")
        opt = stateBits:RemoveFromHash(opt,"AUTO") -- toggling removes auto
        pri = STATE_TOGGLE[pri]
        self.state = bit.bor(pri,opt)
        local key = STATE_VALUE[pri]
        self:SetInitialState(key)
        local t = self:GetAutoFormPause() * 60
        if key == "JOIN" and t > 0 then
            self:Debug("Scheduling UpdateAutoFormState")
            self:ScheduleEvent(self.UpdateAutoFormState,t,self)
        end
        
        self:UpdateText()
        self:UpdateAutoLFG()
    end
end

function lfg:OnDoubleClick(button)
end

function lfg:OnTextUpdate()
    local pri = stateBits:MaskHash(self.state,"PRIMARY")
    self:SetText("LFG:"..self:GetLFGStateShort(STATE_VALUE[pri]).." "..self:GetLFGStatus())
    if not self:IsShowingBusyIcon() then
        self:SetIcon(self:GetLFGStateIcon(STATE_VALUE[pri]))
    end
end

function lfg:GetPriority(name, objective, ...)
    self:Debug("GetPriority",name,objective,...)
    
    return self:GetObjectiveSelection(name,objective)
end

function lfg:SetPriority(name, objective, value, ...)
    self:Debug("SetPriority",name,objective,...)
    
    return self:ToggleObjectiveSelection(name,objective,value)
end

function lfg:FuncPriority(name, func, ...)
    self:Debug("FuncPriority",name, func, ...)
    
    if func == "moveUp" then
        self:ChangePriorityNumber(1,name)
    elseif func == "moveDown" then
        self:ChangePriorityNumber(-1,name)
    elseif func == "update" then
        self:UpdatePriority(nil,name)
    elseif func == "print" then
        func = self.PRIORITY[name] and self.PRIORITY[name][1]
        assert(func,"self.PRIORITY["..name.."][1] == nil")
        
        if string.byte(func) == asterisk then
            func = string.sub(func,2)
        end
        
        assert(self[func],"self["..func.."] == nil")
        
        self:Debug("Executing self[",func,"](self)")
        
        if self:IsDebugging() then
            self:PrintLiteral(self[func](self,false,true))
        end
        
    elseif func == "printall" then
        func = self.PRIORITY[name] and self.PRIORITY[name][1]
        assert(func,"self.PRIORITY["..name.."][1] == nil")
        
        if string.byte(func) == asterisk then
            func = string.sub(func,2)
        end
        
        assert(self[func],"self["..func.."] == nil")
        
        self:Debug("Executing self[",func,"](self,true,true)")
        
        if self:IsDebugging() then
            self:PrintLiteral(self[func](self,true,true))
        end
    end
end

-- =============================================
--             GENERAL EVENTS
-- =============================================

function lfg:OnLogin()
    if self:GetLoginPause() > 0 then
        self.state = stateBits:Get("LOGIN")
        -- TODO schedule state change
    else
        self.state = stateBits:Get(self:GetInitialState())
    end
    
    if not self:GetPriorities() then
        self:LoadLFGProfile("Default")
    end
end

function lfg:UpdateLoginState()
    self:Debug("UpdateLoginState")
    if stateBits:HashContains(self.state,"LOGIN") then
        local initial = self:GetInitialState()
        if initial ~= "DND" then
            self:Print("Auto-LFG enabled. Set",NAME,"to DND to stop looking for a group.")
            PlaySound("PVPENTERQUEUE",NAME,"UpdateLoginState") -- TODO shared sounds?
        end
        self:SetStatus(initial)
    end
end

function lfg:UpdateAutoFormState()
    if stateBits:HashContains(self.state,"JOIN") then
        local pri,opt = stateBits:MaskHash(self.state,"PRIMARY","AUTO")
        self.state = stateBits:AddToHash(opt,"FORM","AUTO")
        self:Print("Trying to form a new group. Disable Auto-Form to stop forming groups after no groups are found.")
        PlaySound("PVPENTERQUEUE",NAME,"UpdateLoginState") -- TODO shared sounds?
        self:UpdateText()
        self:UpdateAutoLFG()
    end
end

function lfg:CHAT_MSG_SYSTEM(arg1,...)
    self:Debug("CHAT_MSG_SYSTEM",arg1,...)
    
    if arg1 == CLEARED_AFK then
        self:UpdateAFKState("") -- do this if states aren't exclusive UnitIsDND("player") and "DND" or "")
        return
    elseif arg1 == CLEARED_DND then
        self:UpdateAFKState("") -- UnitIsAFK("player") and "AFK" or "")
        return
    end
    
    --[[
    ERR_GROUP_FULL
    ERR_GROUP_DISBANDED
    ERR_LEFT_GROUP_YOU
    ERR_RAID_YOU_LEFT
    ERR_UNINVITE_YOU
    ]]
    
    -- TODO should deformat this
    local afk = string.match(arg1,"^You are now ([^%s:]+)")
    if afk == "AFK" then
        self:UpdateAFKState("AFK")
        return
    elseif afk == "DND" then
        self:UpdateAFKState("DND")
        return
    end
end

function lfg:UpdateAFKState(isGoingAway,noupdate)
    self:Debug("UpdateAFKState",isGoingAway)
    if isGoingAway == nil then
        if UnitIsAFK("player") and not self:IsIgnoreAFK() then
            self:Debug("AFK - GOING DND")
            self.state = stateBits:AddToHash(self.state,"AFK")
        elseif UnitIsDND("player") and not self:IsIgnoreDND() then
            self:Debug("DND - GOING DND")
            self.state = stateBits:AddToHash(self.state,"AFK")
        else
            self:Debug("AFK UPDATE - NOT DND")
            self.state = stateBits:RemoveFromHash(self.state,"AFK")
        end
    else
        if isGoingAway == "AFK" and not self:IsIgnoreAFK() then
            self:Debug("AFK event - GOING DND")
            self.state = stateBits:AddToHash(self.state,"AFK")
        elseif isGoingAway == "DND" and not self:IsIgnoreDND() then
            self:Debug("DND event - GOING DND")
            self.state = stateBits:AddToHash(self.state,"AFK")
        else
            self:Debug("AFK/DND event - NOT DND")
            self.state = stateBits:RemoveFromHash(self.state,"AFK")
        end
    end
    
    if not noupdate then
        self:UpdateAutoLFG()
    end
end

function lfg:PARTY_MEMBERS_CHANGED(...)
	self:Debug("PARTY_MEMBERS_CHANGED: ",...)
    
    self:UpdateGroupState()
end

function lfg:RAID_ROSTER_UPDATE(...)
    self:Debug("RAID_ROSTER_UPDATE",...)
    
    self:UpdateGroupState()
end

function lfg:PARTY_LEADER_CHANGED(...)
	self:Debug("PARTY_LEADER_CHANGED",...)
    
    self:UpdateGroupState()
end

function lfg:MEETINGSTONE_CHANGED(...)
	self:Debug("MEETINGSTONE_CHANGED",...)
	-- this event seems to occur when you change your LFG state
    
    self:UpdateLFGState()
end

-- TODO need to update taxi/group/afk/dnd states on login (or soon after)
function lfg:PLAYER_CONTROL_GAINED(...)
    self:Debug("PLAYER_CONTROL_GAINED",...)
    
    if stateBits:HashContains(self.state,"TAXI") then
        self:Debug("scheduling taxi update")
        self:ScheduleEvent(NAME.."TAXI",self.UpdateTaxiState,5,self)
    end
end

function lfg:TAXIMAP_CLOSED(...)
    self:Debug("TAXIMAP_CLOSED",...)
    
    if not self:IsIgnoreTaxi() then
        self:ScheduleEvent(NAME.."TAXI",self.UpdateTaxiState,5,self)
    --    self:UpdateTaxiState()
    end
end

--[[
no need to make functions - inline these
local function IsInRaid()
    return GetNumRaidMembers() > 0
end

local function IsInGroup()
    return GetNumPartyMembers() > 0 or IsPartyLeader()
end

local function IsLeader()
    return IsInRaid() and IsRaidLeader() or IsPartyLeader()
end

local function IsAFK()
    return UnitIsAFK("player")
end

local function IsDND()
    return UnitIsDND("player")
end
]]

function lfg:UpdateGroupState(noupdate)
    self:Debug("UpdateGroupState")
    
    if GetNumRaidMembers() > 0 and IsRaidLeader() or IsPartyLeader() then -- IsLeader()
        self:Debug("GROUP STATE - LEADER")
        self.state = stateBits:AddToHash(self.state,"LEADER")
    elseif GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0 then -- IsInGroup() or IsInRaid()
        self.state = stateBits:RemoveFromHash(self.state,"LEADER")
        if self:IsAvailableInGroup() then
            self.state = stateBits:AddToHash(self.state,"GROUPEDOK")
            self:Debug("GROUP STATE - AVAILABLE")
        else
            self.state = stateBits:AddToHash(self.state,"GROUPED")
            self:Debug("GROUP STATE - DND")
        end
    else
        self:Debug("GROUP STATE - AVAILABLE")
        local grp,etc = stateBits:MaskHash(self.state,"GROUPED","GROUPEDOK","LEADER")
        if grp ~= 0 then
            self:Debug("LEFT GROUP")
            
            -- special left group transition
            local pri,etc = stateBits:MaskHash(etc,"FORM")
            if pri ~= 0 then
                self:Debug("LEFT GROUP WHEN LFM")
                pri = STATE_TOGGLE[pri] -- FORM -> AVAILABLE
                etc = bit.bor(pri,etc)
            end
            
            -- odd behavior when reloading and group leader
            self.state = etc
            self:UpdateText()
        end
    end
    
    if not noupdate then
        self:UpdateAutoLFG()
    end
    
    
    -- TODO consider capturing the CHAT_MSG_SYSTEM events you join/leave the group or disbanded messages
    --[[ todo IsPartyLeader() too IsRaidLeader(), IsRaidOfficer(), GetNumPartyMembers()
    
    GetNumRaidMembers()
    ]]
end

function lfg:UpdateTaxiState(noupdate)
    if UnitOnTaxi("player") and not self:IsIgnoreTaxi() then
        self:Debug("TAXI STATE - ON TAXI")
        self.state = stateBits:AddToHash(self.state,"TAXI")
    else
        self:Debug("TAXI STATE - OFF TAXI")
        self.state = stateBits:RemoveFromHash(self.state,"TAXI")
    end
    
    if not noupdate then
        OnLFGUpdate(self) -- update LFG immediately (already scheduled one delay event to update taxi)
    end
    
    --[[
    UnitOnTaxi("unit")
    PLAYER_CONTROL_LOST
    TAXIMAP_CLOSED
    ]]
end

function lfg:UpdateLFGState()
    self:Debug(GetLookingForGroup())
    
    -- Get LFG status
    local type1, name1, type2, name2, type3, name3,
        lfmType, lfmName, comment, queued, lfgStatus,
        lfmStatus, autoaddStatus = GetLookingForGroup()

    if self.categories[type1] == "None" then
        self.lfg1 = nil
    else
        self.lfg1 = type1..';'..name1
    end
    
    if self.categories[type2] == "None" then
        self.lfg2 = nil
    else
        self.lfg2 = type2..';'..name2
    end
    
    if self.categories[type3] == "None" then
        self.lfg3 = nil
    else
        self.lfg3 = type3..';'..name3
    end
    
    if self.categories[lfmType] == "None" then
        self.lfm = nil
    else
        self.lfm = lfmType..';'..lfmName
    end
    
    local c,o = self:GetQueryText(self.lfg1)
    self:Debug("LFG1",c or "None",c and o or "")
    c,o = self:GetQueryText(self.lfg2)
    self:Debug("LFG2",c or "None",c and o or "")
    c,o = self:GetQueryText(self.lfg3)
    self:Debug("LFG3",c or "None",c and o or "")
    c,o = self:GetQueryText(self.lfm)
    self:Debug("LFM",c or "None",c and o or "")
    
    self:UpdateText()
end


-- =============================================
--           FILTER DATA FUNCTIONS
-- =============================================

-- inspired by color function in tourist lib
-- quest level is "min" and max is +4
function lfg:GetLevelColor(objectiveLevel)
    local playerLevel = UnitLevel("player") -- TODO is this too costly? should it be cached?
    -- TODO should base this on GetQuestGreenRange()
    
	if playerLevel == objectiveLevel then
		return 1, 1, 0
	elseif playerLevel <= objectiveLevel - 4 then -- 4 levels is the +hit penalty threshold
		return 1, 0, 0
	elseif playerLevel <= objectiveLevel then -- orangeish
		return 1, (playerLevel - objectiveLevel - 4) / -8, 0
	elseif playerLevel <= objectiveLevel + 3 then -- greenish
		return 2 * (playerLevel - objectiveLevel) / 3, 1, 0
	elseif playerLevel <= objectiveLevel + 6 then -- greyish
		local num = (playerLevel - objectiveLevel) / 6
		return num, 1 - num, num
	else
		return 0.5, 0.5, 0.5
	end
end

function lfg:GetGroupSize(query,category,objective)
    local o
    if not query then
        query = self.criteria[objective]
        assert(query,"unknown objective: "..objective)
    end
    
    if not category then
        category,o = self:GetQueryText(query)
        assert(category,"unknown query: "..query)
    end
    
    if category == "Zone" then
        return 5
    elseif category == "None" then
        return 1
    elseif category == "Heroic Dungeon" then
        return 5
    elseif category == "Dungeon" then
        return 5
    end
    
    if not objective then
        objective = o or self:GetObjectiveText(query)
    end
    
    if category == "Raid" or category == "Battleground" then
        return self:GetInstanceGroupSize(objective)
    else
        return self:GetQuestGroupSize(objective)
    end
end

-- =============================================
--             LFG DATA FUNCTIONS
-- =============================================

function lfg:GetLFGStatus()
    -- players in your LFG pool, players in all your LFG preferences, all players LFG
    return string.format("(%d/%d/%d)%s",0,0,self.whoCount,self:GetQuery() and "?" or "")
end

function lfg:GetWhoData(name)
    local who = self.who[name]
    if not who then return end
    self:Debug("who",name,who)
    return string.match(who,string.rep("([^;]*);?",10))
end

function lfg:GetPartyData(name)
    local party = self.partyWho[name]
    if not party then return end
    self:Debug("party",name,party)
    
    return string.match(party,string.rep("([^,]*),?([^,]*),?([^;]*);?",10))
end

function lfg:UPDATE_LFG_TYPES(...)
	self:Debug("UPDATE_LFG_TYPES: ",...)
    
    self:CancelScheduledEvent("UPDATE_LFG_TYPES") -- cancel if we recieve a natural update
	
    packInto(self.categories,Sys_GetLFGTypes())
    
    self:OnCategoryUpdate(self.categories)
    
    -- TODO throttle this?
    self:UpdateObjectives()
end

function lfg:UpdateFilters()
    -- TODO throttle this?
    for k,v in pairs(self.OnFilterUpdate) do
        local ret, err = pcall(v,self)
        if not ret then
            error(k.." - "..err)
        end
    end
end

function lfg:UpdateObjectives()
    self:ClearObjectives()
    
	local objectives,count,lookupTable = self.objectives
	
	for catNum,category in pairs(self.categories) do
        if category == "Zone" then
            lookupTable = self.zones
        elseif string.find(category,"Quest",1,true) then
            lookupTable = self.quests
        elseif category == "Battleground" then
            lookupTable = self.battlegrounds
        elseif category == "Dungeon" then
            lookupTable = self.instances
            -- dungeons (and raids?) show up like "Hellfire Citadel - Blood Furnace"
            -- DOES NOT always match the zone name (gee thx blizz)
        elseif category == "Heroic Dungeon" then
            lookupTable = self.heroic
            -- heroic dungeons show up like "Hellfire Citadel - Blood Furnace (Heroic)"
        elseif category == "Raid" then
            lookupTable = self.instances
        else
            lookupTable = nil
        end
        
		count = self:AddObjectives(objectives,lookupTable,self.criteria,catNum,Sys_GetLFGTypeEntries(catNum))
        if not self.categoryCache[catNum] then
            self.categoryCache[catNum] = {0,0,1}
        end
        
        self.categoryCache[catNum][3] = count
	end
    
    if self:IsDebugging() then
        self:PrintLiteral(self.objectives)
    end
    
    for k,v in pairs(self.OnObjectivesUpdate) do
        local ret, err = pcall(v,self)
        if not ret then
            error(k.." - "..err)
        end
    end
end

function lfg:LFG_UPDATE(...)
    -- what does this indicate? a return of the query to your available objectives?
    -- not sure... it doesn't fire when you change your LFG state
	self:Debug("LFG_UPDATE: ",...)
    
    self:UpdateLFGState()
end

function lfg:ClearObjectives()
    for k,v in pairs(self.objectives) do
        self.objectives[k] = nil
    end
end

function lfg:AddObjectives(objectives,lookupTable1,lookupTable2,catNum,...)
    self:Debug("add objectives ",catNum,...)
    
    local n = select("#",...)
	for i=1, n, 2 do
        local k = select(i,...)
        local v = select(i+1,...)
        local query
--        if string.len(v) < 1 then
            query = catNum..';'..((i+1)/2)
            objectives[query] = k
            self:Debug("Objective: ",k)
            -- TODO what are these used for? they don't constitute valid query
--[[        elseif k ~= v then
            self:Debug(k,"=",v)
            objectives[catNum..';'..((i+1)/2)] = k
            self:Debug("Objective: ",k," = ",v)
        else
            objectives[catNum..';'..v] = k
            self:Debug("Objective: ",k," = ",v)]]
--        end
        if lookupTable1 then
            lookupTable1[k] = query
        end
        
        if lookupTable2 then
            lookupTable2[k] = query
        end
	end
    
    return n / 2
end

function lfg:GetQueryText(query)
    return self.categories[tonumber((Split(query,';')))], self.objectives[query]
end

function lfg:GetCategoryText(number)
    assert(type(number) == "number")
    return self.categories[number]
end

function lfg:GetObjectiveText(query)
    return self.objectives[query]
end

function lfg:PrintAvailable()
    self:Debug("PrintAvailable")
    local list,a = {}
    for query,objText in pairs(self.objectives) do
        a = self.categories[tonumber((Split(query,';')))] or '?'
        if not list[a] then
            list[a] = {}
        end
        table.insert(list[a],objText)
    end
    
    for category, objectives in pairs(list) do
        for i,objText in ipairs(objectives) do
            self:CustomPrint(1,1,0,nil,nil,":",category,objText)
        end
    end
end

-- =============================================
--             AUTO LFG FUNCTIONS
-- =============================================

-- TODO consider making auto-lfg a module

--[[
    "DND",          -- player selected DND
    "AVAILABLE",    -- player selected available
    "JOIN",         -- player selected join
    "FORM",         -- player selected form
    
    "AUTO",         -- auto upgrade/downgrade of state
    "TAXI",         -- player on a taxi, will transition to primary state
    "GROUPED",      -- player is grouped, will transition to available or dnd
    "AFK"           -- player is <AFK> or <DND> and preferences recognize these
]]

local function SelectLFGObjectives(self,t,max,...)
    self:Debug("SelectLFGObjectives",t,max,...)
    for i=2,select('#',...),2 do
        if #t >= max then
            return
        end
        
        local obj = select(i-1,...)
        
        -- TODO need to cycle LFG objectives periodically
        if "OK" == select(i,...) then
            table.insert(t,obj)
        end
    end
end

local tempLFG = {}
local function SelectLFGPriorities(self)
    while #tempLFG > 0 do
        table.remove(tempLFG)
    end
    
    for i,v in ipairs(self:GetPriorities()) do
        if #tempLFG >= 3 then
            return unpack(tempLFG)
        end
        
        if v == self.PRIORITY_IGNORE_HEADER then
            return unpack(tempLFG)
        end
        
        local func = self.PRIORITY[v] and self.PRIORITY[v][1]
        self:Debug("examining",v,func)
        if func then
            if string.byte(func) == asterisk then
                -- TODO selection needs to respect filters
                SelectLFGObjectives(self,tempLFG,3,self:ListObjectiveSelection(name))
            else
                SelectLFGObjectives(self,tempLFG,3,self[func](self))
            end
        end
    end
    
    return unpack(tempLFG)
end

function OnLFGUpdate(self)
    local pri,opt = stateBits:MaskHash(self.state,"PRIMARY")
    
    -- TODO deal with complex states & transitions
    -- DND
    --  compatible with all modifiers
    -- AVAILABLE
    --  suppressed by GROUPED
    --  suppressed by AFK
    --  partially suppressed by TAXI
    -- JOIN
    --  suppressed by GROUPED (unless LEADER)
    --  GROUPEDOK transitions to AVAILABLE
    --  LEADER transitions to FORM
    --  partially suppressed by TAXI
    --  suppressed by AFK
    -- FORM
    --  suppressed by GROUPED (unless LEADER)
    --  GROUPEDOK transitions to AVAILABLE
    --  suppressed by AFK
    
    if pri == DND then
        self:Debug("DND - SETTING - LFG OFF")
        return
    end
    
    if stateBits:HashContains(opt,"AFK") then
        self:Debug("AFK - SETTING - LFG OFF")
        -- TODO suppressed icon/text
        return
    end
    
    if stateBits:HashContains(opt,"LEADER") and pri == JOIN then
        pri = FORM
        self.state = bit.bor(pri,opt)
        self:UpdateText()
    elseif stateBits:HashContains(opt,"GROUPEDOK") and pri ~= AVAILABLE then
        pri = AVAILABLE
        self.state = bit.bor(pri,opt)
        self:UpdateText()
    elseif stateBits:HashContains(opt,"GROUPED") then
        self:Debug("GROUPED - SETTING - LFG OFF")
        -- TODO suppressed icon/text
        return
    end
    
    if pri == FORM then
        self:Print("SETTING - LFM",self.lfg1)
        -- TODO get lfg1
        return
    end
    
    local taxi = stateBits:HashContains(opt,"TAXI")
    
    if taxi then
        self:Debug("ON TAXI - FILTERING PRIORITIES")
        -- TODO suppressed icon/text
    end
    
    local autojoin = pri == JOIN
    
    if autojoin then
        self:Debug("SETTING - AUTOJOIN")
    else
        assert(pri == AVAILABLE)
    end
    
    self:Debug("SETTING - LFG")
    
    local lfg1, lfg2, lfg3 = SelectLFGPriorities(self)
    
    self:Print("SELECTED LFG OBJECTIVES",lfg1,lfg2,lfg3)
    lfg1,lfg2,lfg3 = lfg1 and self.criteria[lfg1],lfg2 and self.criteria[lfg2],lfg3 and self.criteria[lfg3]
    self:Print("LFG QUERIES",lfg1,lfg2,lfg3)
end

function lfg:UpdateAutoLFG(delay)
    self:ScheduleEvent(NAME.."LFG_UPDATE",OnLFGUpdate,delay or 1,self)
end

function lfg:IsObjectiveFiltered(catNum, objText, query)
    if objText then
        if not query then
            query = self.criteria[objText]
            assert(query,"Could not find query for objective: "..objText)
        end
        
        if not catNum then
            catNum = tonumber(Split(query))
            assert(catNum,"Count not find category number in query: "..query)
        end
    else
        assert(query,"arg #2 (objective text) or arg #3 (query) is required")
        
        if not catNum then
            catNum = tonumber(Split(query))
            assert(catNum,"Count not find category number in query: "..query)
        end
        
        if not objText then
            objText = self:GetObjectiveText(query)
            assert(objText,"Could not find objective text for query: "..query)
        end
    end
    
    -- TODO filter these
end

-- =============================================
--             LFG WHO FUNCTIONS
-- =============================================

-- TODO consider making lfgwho a module

--[[
local quot = string.byte('"')
local dash = string.byte('-')
local space = string.byte(' ')
local whoPattern = '[^"%s]*%b""'
local levelPattern = '(%d*)%-(%d*)'


BROKEN BROKEN BROKEN

how can I use %b"" to OPTIONALLY matched balanced spaces?

local function GetWhoIter(text)
    return WhoIter,text,nil --string.gmatch(text,whoPattern)
end

local function WhoParse(conditions,f,s,var,...)
    var = f(s,var)
    
    if var == nil then
        -- end recursion
        io.stdout:write("return\n")
        return conditions,...
    end
    io.stdout:write(var or "nil","\n")
    
    local isLevel, min, max = string.match(var,levelPattern)
    if string.len(isLevel) > 0 then
        io.stdout:write("level '",isLevel,"' '",min or "nil","' '",max or "nil","'\n")
        
        -- level match
        conditions["min"] = min and string.len(min) > 0 and tonumber(min) or -1
        conditions["max"] = max and string.len(max) > 0 and tonumber(max) or 9999
    elseif string.byte(var,2) ~= dash then
        io.stdout:write("raw\n")
        
        -- var twice, 1st = var, 2nd adding condition to varargs
        return WhoParse(conditions,f,s,var,var,...)
    elseif string.byte(var,3) == quot then
        io.stdout:write("quot\n")
        
        conditions[string.sub(var,1,1)] = string.sub(var,4,-2)
    else
        io.stdout:write("unquot\n")
    
        conditions[string.sub(var,1,1)] = string.sub(var,3)
    end
    
    return WhoParse(conditions,f,s,var,...)
end

local function WhoConditions(text)
    return WhoParse({},GetWhoIter(text))
end

local cond,raw1 = WhoConditions('z-"on Fo" n-"Ual" 10-30 raw')

assert(raw1 == "raw","Got "..(raw1 or "nil"))
assert(cond["min"] == 10,"Got "..(cond["min"] or "nil"))
assert(cond["max"] == 30,"Got "..(cond["max"] or "nil"))
assert(cond["z"] == "on Fo","Got "..(cond["z"] or "nil"))
assert(cond["n"] == "Ual","Got "..(cond["n"] or "nil"))
]]

function lfg:LFGWho(text)
    self:Debug("LFGWho(",text,")")
    
    --[[
    local raw = {WhoConditions(text)}
    local conditions = table.remove(raw,1)
    
    if self:IsDebugging() then
        self:PrintLiteral("LFGWho",conditions)
    end
    ]]
    
    
    local catNum = nil
    local t = time()
    for query,objText in pairs(self.objectives) do
        self:AddQuery(t,query)
    end
    
    self:ProcessQuery(false)
end

function lfg:GetQuery(nosearch)
    local t,query
	for k,v in pairs(self.queue) do
		if not t or v < t then
			t = v
			query = k
		end
        
		if nosearch then return t,query end
	end
    
    return t,query
end

function lfg:RemoveQuery(query)
    self.queue[query] = nil
end

function lfg:AddQuery(t,query)
	self.queue[query] = t
    self.lastQuery = t
end

function lfg:ProcessQuery(search)
    self:Debug("ProcessQuery",search)
    local now = time()
	
    local t,query
    if type(search) == "string" then
        t,query = now,search
        if self.currentQuery then
            self:AddQuery(now,query)
            return
        end
    else
        t,query = self:GetQuery(not search)
    end
    
    if self.currentQuery and self.lastQuery - 10 > now then
        self:Debug("throttling query")
        return
    end
    
	if not query then
        self:Debug("no query found")
        return
    end
    
    if t > now then
        self:Debug("Next query at",t)
        -- loop protection
--        self:ScheduleEvent(self.ProcessQuery,1,self,true)
        return
    end
    
	local category,objective = Split(query,';')
    
    self:Debug("ProcessQuery",category,objective)
    
    if not CanSendLFGQuery(category,objective) then
        self:Debug("Can't Query",category,objective)
        self:AddQuery(now + 1,query)
        self:ProcessQuery(true)
    else
        LFGQuery(category,objective)
        -- calling another LFGQuery in succession wipes out the previous query apparently
    end
end

function AddToCache(cache, fromIndex, name)
    local i,v
    cache[1] = t -- TODO move this to outside the add/rem function?
    for i=fromIndex, #cache do
        v = cache[i]
        
        if name == v then
            return false
        end
        
        if name > v then
            table.insert(cache,i,name)
            return true
        end
    end
    table.insert(cache,name)
    return true
end

function RemoveFromCache(cache, fromIndex, name)
    for i=fromIndex, #cache do
        if cache[i] == name then
            table.remove(cache,i)
            return true
        end
    end
    return false
end

-- returns categoryNumber,criteria1,criteria2,criteria3
function GetRemovedCategoryAndCriteria(self,query,criteria1,criteria2,criteria3)
        local query1, query2, query3 =
            self.criteria[criteria1],self.criteria[criteria2],self.criteria[criteria3]
        local category = Split(query,';')
        local remCat
        
        self:Debug("examining criteria",criteria1,criteria2,criteria3)
        self:Debug("looking to remove criteria",query1,query2,query3)
        
        if query == query3 then
            criteria3 = ""
            if Split(query2,';') ~= category and Split(query1,';') ~= category then
                remCat = tonumber(category)
            end
        end
        
        if query == query2 then
            criteria2 = ""
            if Split(query1,';') ~= category then
                remCat = tonumber(category)
            end
        end
        
        if query == query1 then
            criteria1 = ""
            remCat = tonumber(category)
        end
        
        return remCat,criteria1,criteria2,criteria3
end

local tempOldCache = {}

local function storeLFGParty(self,...)
    if select('#',...) < 3 then return end
    
    local n = select('#',...) / 3
    
    return string.format(string.rep("%s,%s,%s;",n),...)
end

local function GetLFGParty(category, objective, i, num)
    if num < 1 then
        return
    end
    
    local name, level, class = GetLFGPartyResults(category, objective, i, num)
    
    return name, level, class, GetLFGParty(category, objective, i, num-1)
end

function lfg:UPDATE_LFG_LIST(...)
    local query = self.currentQuery
    
    if not query then
        return
    end
    
    local category, objective = Split(query,';')
    category = tonumber(category)
    local catText,objText = self:GetQueryText(query)
    local count = GetNumLFGResults(category,objective)
    
	self:Debug("UPDATE_LFG_LIST",category or "?",objective or "?",count,"args:",...)
    
    local t = time()
    
    self:ClearCache(t, category, query, tempOldCache)
    
    local name, numPartyMembers
    for i=1,count do
        if self:IsDebugging() then
            self:PrintLiteral(GetLFGResults(category, objective, i))
        end
        name, _, _, _, _, _, _, _, _, numPartyMembers, _
            = self:AddPlayer(t,GetLFGResults(category, objective, i))
        
        -- TODO add party members
        -- more than 10 members total
--        if numPartyMembers > 9 then
--            -- do raid listing
--            -- format(LFM_NUM_RAID_MEMBER_TEMPLATE,numPartyMembers)
--        else
        if numPartyMembers > 0 and numPartyMembers <= 9 then
            self.partyWho[name] = storeLFGParty(self,GetLFGParty(category, objective, i, numPartyMembers))
        end
        
        AddToCache(self.cache[query],2,name)
        AddToCache(self.categoryCache[category],4,name)
        
        -- remove from old table
        tempOldCache[name] = nil
    end
    
    if count > 0 then
        self.cache[query][1] = t
        self.categoryCache[category][1] = t
    end
    
    if self:IsDebugging() then
        self:PrintLiteral("old cache",tempOldCache)
    end
    
    for name,v in pairs(tempOldCache) do
        -- remove criteria from who data
        local time, level, zone, class, criteria1,
            criteria2, criteria3, comment, numParty, isLFM = self:GetWhoData(name)
        
        local catNum,criteria1,criteria2,criteria3 =
            GetRemovedCategoryAndCriteria(self,query,criteria1,criteria2,criteria3)
        
        if catNum then
            self:Debug("removing category",catNum)
            RemoveFromCache(self.categoryCache[catNum],4,name)
        end
        
        if criteria3 == "" and criteria2 == "" and criteria1 == "" then
            -- if no criteria left remove who data
            self:Debug("removing",name)
            self:RemovePlayer(name)
        else
            self:AddPlayer(t, name, level, zone, class, criteria1,
                criteria2, criteria3, comment, numParty, isLFM)
        end
        
        -- TODO also remove from categoryCache depending on conditions present
        
        tempOldCache[name] = nil
    end
    
    -- clear temp table
    for k,v in pairs(tempOldCache) do
        tempOldCache[k] = nil
    end
    
    self.currentQuery = nil
     
    local hasQuery = self:GetQuery()
    if hasQuery and not self:IsBusyIcon() then
        self:Print("Grouper needs a user click!")
        self:StartBusyIcon()
    elseif self:IsBusyIcon() and not hasQuery then
        self:Print("Thanks, you can stop clicking now.")
        self:CancelBusyIcon()
    end
--    self:ProcessQuery(true) -- PROTECTED (in non-interactive scope)
--    self:ScheduleEvent(self.ProcessQuery,1,self,true)
    
    self:UpdateText()
    self:OnLFGUpdate(catText,query)
    self:OnObjectiveCountUpdate(catText,query)
    self:OnCategoryCountUpdate(category)
end

function lfg:RemovePlayer(name)
    if self.who[name] then
        self.whoCount = self.whoCount - 1
    end
    
    self.who[name] = nil
end

function lfg:AddPlayer(time, name, level, zone, class, criteria1, criteria2, criteria3, comment, numPartyMembers, isLFM)
    if self:IsDebugging() then
        self:CustomPrint(1,1,0,nil,nil," ",name,"Level",level,class,"-",zone,
            criteria1 and ("LFG: "..criteria1) or "", criteria2 or "", criteria2 or "",
            comment and string.len(comment) > 0 and ("Comment: "..comment) or "",
            numPartyMembers and tonumber(numPartyMembers) > 0 and ("Party: "..numPartyMembers) or "")
    end
    
    if not self.who[name] then
        self.whoCount = self.whoCount + 1
    end
    
    self.who[name] = string.format("%d;%d;%s;%s;%s;%s;%s;%s;%d;%s", time, level, zone,
        class or "?", criteria1 or "", criteria2 or "", criteria3 or "", comment or "",
        numPartyMembers or 0, tostring(isLFM and true or false))
        
    return name, time, level, zone, class, criteria1, criteria2, criteria3, comment, numPartyMembers, isLFM
end

function lfg:ClearCache(t, category, query, oldmap)
    local cache
    cache,self.cache[query] = self.cache[query],{t}
    
    assert(type(category) == "number")
    
    if not cache then
        self.categoryCache[category][2] = self.categoryCache[category][2] + 1
    end
    self.categoryCache[category][1] = t
    
    if cache and oldmap and type(oldmap) == "table" then
        for i,v in ipairs(cache) do
            if i ~= 1 then
                oldmap[v] = i
            end
        end
    end
    
    return cache
end

-- text,count,time,scanned,total
function lfg:GetCategoryStatus(category)
    assert(type(category) == "number")
    local cache = self.categoryCache[category]
    if not cache then
        return "?",0,0,0,1
    end
    local count = #cache-3
    
    return (cache[2] > 0 and count or "?")..(cache[2] == cache[3] and "" or "+"),
        count, cache[1], cache[2], cache[3]
end

-- text,count,time
function lfg:GetObjectiveStatus(query)
    local cache = self.cache[query]
    if not cache then
        return "?",0,0
    end
    local count = #cache-1
    
    return (cache[1] > 0 and tostring(count) or "?"),count,cache[1]
end